"""
Test the wntr.network.elements classes
"""

import pytest
import unittest
from os.path import abspath, dirname, join
from unittest import SkipTest

import pytest
import numpy as np
import wntr
import wntr.network.elements as elements
from wntr.network.options import TimeOptions
from wntr.utils.check_values import _check_float, _check_float_or_none, \
    _check_positive_non_zero_float, _check_positive_or_zero_float

testdir = dirname(abspath(str(__file__)))
datadir = join(testdir, "networks_for_testing")
net1dir = join(testdir, "..", "..", "examples", "networks")


class TestElements(unittest.TestCase):
    def test_Curve(self):
        pts1 = [[3, 5]]
        pts2 = [[3, 6], [7, 2], [10, 0]]
        pts3 = [[3, 6], [7, 2], [10, 1]]
        expected_str = (
            "<Curve: 'curve2', curve_type='HEAD', points=[[3, 6], [7, 2], [10, 0]]>"
        )
        # Create the curves
        curve1 = elements.Curve("curve1", "PUMP", pts1)
        curve2a = elements.Curve("curve2", "HEAD", pts2)
        curve2b = elements.Curve("curve2", "HEAD", pts3)
        curve2c = elements.Curve("curve2", "HEAD", pts3)
        # Test that the assignments are working
        self.assertListEqual(curve2b.points, pts3)
        self.assertEqual(curve1.num_points, 1)
        self.assertEqual(len(curve2c), 3)
        # Testing __eq__
        self.assertNotEqual(curve1, curve2a)
        self.assertNotEqual(curve2a, curve2b)
        self.assertEqual(curve2b, curve2c)
        # testing __getitem__ and __getslice__
        self.assertListEqual(curve2a[0], [3, 6])
        self.assertListEqual(curve2a[:2], [[3, 6], [7, 2]])
        # verify that the points are being deep copied
        self.assertNotEqual(id(curve2b.points), id(curve2c.points))

    def test_Pattern(self):
        pattern_points1 = [1, 2, 3, 4, 3, 2, 1]
        pattern_points2 = [1.0, 1.2, 1.0]
        pattern_points3 = 3.2

        # test constant pattern creation
        timing1 = TimeOptions()
        timing1.pattern_start = 0
        timing1.pattern_timestep = 1

        timing2 = TimeOptions()
        timing2.pattern_start = 0
        timing2.pattern_timestep = 5
        pattern1a = elements.Pattern(
            "constant", multipliers=pattern_points3, time_options=timing1
        )
        pattern1b = elements.Pattern(
            "constant", multipliers=[pattern_points3], time_options=(0, 1)
        )
        self.assertListEqual(pattern1a.multipliers.tolist(), [pattern_points3])
        self.assertTrue(np.all(pattern1a.multipliers == pattern1b.multipliers))
        self.assertFalse(id(pattern1a.multipliers) == id(pattern1b.multipliers))
        self.assertEqual(pattern1a.time_options, pattern1b.time_options)

        # def multipliers setter
        pattern2a = elements.Pattern(
            "oops", multipliers=pattern_points3, time_options=(0, 5)
        )
        pattern2b = elements.Pattern(
            "oops", multipliers=pattern_points1, time_options=timing2
        )
        pattern2a.multipliers = pattern_points1
        self.assertEqual(pattern2a.time_options, pattern2b.time_options)

        # test pattern evaluations
        expected_value = pattern_points1[2]
        self.assertEqual(pattern2a[2], expected_value)
        self.assertEqual(pattern2b.at(10), expected_value)
        self.assertEqual(pattern2b.at(12.5), expected_value)
        self.assertEqual(pattern2b.at(14), expected_value)
        self.assertEqual(pattern2b.at(9 * 5), expected_value)
        self.assertNotEqual(pattern2b.at(15), expected_value)

        pattern3 = elements.Pattern(
            "nowrap", multipliers=pattern_points2, time_options=(0, 100), wrap=False
        )
        self.assertEqual(pattern3[5], 0.0)
        self.assertEqual(pattern3[-39], 0.0)
        self.assertEqual(pattern3.at(-39), 0.0)
        self.assertEqual(pattern3.at(50), 1.0)

        pattern4 = elements.Pattern("constant")
        self.assertEqual(len(pattern4), 0)
        self.assertEqual(pattern4.at(492), 1.0)

        pattern5a = elements.Pattern(
            "binary",
            [0.0, 0.0, 1.0, 1.0, 1.0, 1.0, 0.0, 0.0, 0.0],
            time_options=(0, 1),
            wrap=False,
        )
        pattern5b = elements.Pattern.binary_pattern(
            "binary", step_size=1, start_time=2, end_time=6, duration=9
        )
        self.assertFalse(pattern5a.__eq__(pattern5b))
        self.assertTrue(
            np.all(np.abs(pattern5a.multipliers - pattern5b.multipliers) < 1.0e-10)
        )

    def test_pattern_interpolation(self):
        timing = TimeOptions()
        timing.pattern_interpolation = True
        p = elements.Pattern('p1', multipliers=[1, 1.2, 1.6], time_options=timing)
        self.assertAlmostEqual(p.at(0), 1)
        self.assertAlmostEqual(p.at(3600), 1.2)
        self.assertAlmostEqual(p.at(7200), 1.6)
        self.assertAlmostEqual(p.at(1800), 1.1)
        self.assertAlmostEqual(p.at(5400), 1.4)
        self.assertAlmostEqual(p.at(2250), 0.2/3600*2250 + 1)
        self.assertAlmostEqual(p.at(9000), 1.3)
        self.assertAlmostEqual(p.at(12600), 1.1)

    def test_TimeSeries(self):
        wn = wntr.network.WaterNetworkModel()

        pattern_points2 = [1.0, 1.2, 1.0]
        wn.add_pattern("oops", pattern_points2)
        pattern2 = wn.get_pattern("oops")
        pattern5 = elements.Pattern.binary_pattern(
            "binary", step_size=1, start_time=2, end_time=6, duration=9
        )
        wn.add_pattern("binary", pattern5)
        base1 = 2.0

        # test constructor and setters, getters
        tvv1 = elements.TimeSeries(wn.patterns, base1, None, None)
        tvv2 = elements.TimeSeries(wn.patterns, base1, "oops", "tvv2")
        self.assertRaises(ValueError, elements.TimeSeries, *("A", None, None))
        self.assertRaises(ValueError, elements.TimeSeries, *(1.0, "A", None))
        self.assertEqual(tvv1._base, base1)
        self.assertEqual(tvv1.base_value, tvv1._base)
        self.assertEqual(tvv1.pattern_name, None)
        self.assertEqual(tvv1.pattern, None)
        self.assertEqual(tvv1.category, None)
        tvv1.base_value = 3.0
        self.assertEqual(tvv1.base_value, 3.0)
        tvv1.pattern_name = "binary"
        self.assertEqual(tvv1.pattern_name, "binary")
        tvv1.category = "binary"
        self.assertEqual(tvv1.category, "binary")

        # Test getitem
        # print(tvv1)
        # print(tvv2, pattern2)
        self.assertEqual(tvv1.at(1), 0.0)
        self.assertEqual(tvv1.at(7202), 3.0)
        self.assertEqual(tvv2.at(1), 2.0)
        # print(tvv2, pattern2.time_options)
        self.assertEqual(tvv2.at(3602), 2.4)

        price1 = elements.TimeSeries(wn.patterns, 35.0, None)
        price2 = elements.TimeSeries(wn.patterns, 35.0, None)
        self.assertEqual(price1, price2)
        self.assertEqual(price1.base_value, 35.0)

        speed1 = elements.TimeSeries(wn.patterns, 35.0, pattern5)
        speed2 = elements.TimeSeries(wn.patterns, 35.0, pattern5)
        self.assertEqual(speed1, speed2)
        self.assertEqual(speed1.base_value, 35.0)

        head1 = elements.TimeSeries(wn.patterns, 35.0, pattern2)
        head2 = elements.TimeSeries(wn.patterns, 35.0, pattern2)
        self.assertEqual(head1, head2)
        self.assertEqual(head1.base_value, 35.0)

        demand1 = elements.TimeSeries(wn.patterns, 1.35, pattern2)
        demand2 = elements.TimeSeries(wn.patterns, 1.35, pattern2)
        self.assertEqual(demand1, demand2)
        self.assertEqual(demand1.base_value, 1.35)

    #    expected_values1 = np.array([1.35, 1.62, 1.35, 1.35, 1.62])
    #    demand_values1 = demand2.get_values(0, 40, 10)
    #    self.assertTrue(np.all(np.abs(expected_values1-demand_values1)<1.0e-10))
    #    expected_values1 = np.array([1.35, 1.35, 1.62, 1.62, 1.35, 1.35, 1.35, 1.35, 1.62])
    #    demand_values1 = demand2.get_values(0, 40, 5)
    #    self.assertTrue(np.all(np.abs(expected_values1-demand_values1)<1.0e-10))
    #
    #    source1 = elements.Source('source1', 'NODE-1131', 'CONCEN', 1000.0, pattern5)
    #    source2 = elements.Source('source1', 'NODE-1131', 'CONCEN', 1000.0, pattern5)
    #    self.assertEqual(source1, source2)
    #    self.assertEqual(source1.strength_timeseries.base_value, 1000.0)

    def test_Demands(self):
        wn = wntr.network.WaterNetworkModel()

        pattern_points1 = [0.5, 1.0, 0.4, 0.2]
        pattern1 = elements.Pattern(
            "1", multipliers=pattern_points1, time_options=(0, 10)
        )
        pattern_points2 = [1.0, 1.2, 1.0]
        pattern2 = elements.Pattern(
            "2", multipliers=pattern_points2, time_options=(0, 10)
        )
        demand1 = elements.TimeSeries(wn.patterns, 2.5, pattern1, "_base_demand")
        demand2 = elements.TimeSeries(wn.patterns, 1.0, pattern2, "residential")
        demand3 = elements.TimeSeries(wn.patterns, 0.8, pattern2, "residential")

        expected1 = 2.5 * np.array(pattern_points1 * 3)
        expected2 = 1.0 * np.array(pattern_points2 * 4)
        expected3 = 0.8 * np.array(pattern_points2 * 4)
        expectedtotal = expected1 + expected2 + expected3
        expectedresidential = expected2 + expected3
        demandlist1 = elements.Demands(wn.patterns, demand1, demand2, demand3)
        demandlist2 = elements.Demands(wn.patterns)
        demandlist2.append(demand1)
        demandlist2.append(demand1)
        demandlist2[1] = demand2
        demandlist2.append((0.8, pattern2, "residential"))
        self.assertListEqual(list(demandlist1), list(demandlist2))

        demandlist2.extend(demandlist1)
        self.assertEqual(len(demandlist1), 3)
        self.assertEqual(len(demandlist2), 6)
        del demandlist2[3]
        del demandlist2[3]
        del demandlist2[3]
        del demandlist2[0]
        demandlist2.insert(0, demand1)
        self.assertListEqual(list(demandlist1), list(demandlist2))
        demandlist2.clear()
        self.assertEqual(len(demandlist2), 0)
        self.assertFalse(demandlist2)

        raise SkipTest

        self.assertEqual(demandlist1.at(5), expectedtotal[0])
        self.assertEqual(demandlist1.at(13), expectedtotal[1])
        self.assertEqual(demandlist1.at(13, "residential"), expectedresidential[1])
        self.assertTrue(
            np.all(np.abs(demandlist1.get_values(0, 110, 10) - expectedtotal) < 1.0e-10)
        )
        self.assertListEqual(demandlist1.base_demand_list(), [2.5, 1.0, 0.8])
        self.assertListEqual(demandlist1.base_demand_list("_base_demand"), [2.5])
        self.assertListEqual(demandlist1.pattern_list(), [pattern1, pattern2, pattern2])
        self.assertListEqual(
            demandlist1.pattern_list(category="residential"), [pattern2, pattern2]
        )
        self.assertListEqual(
            demandlist1.category_list(), ["_base_demand", "residential", "residential"]
        )

    def test_fire_fighting_demand(self):
        # Setup network
        wn = wntr.network.WaterNetworkModel()
        wn.add_junction(
            "new_junction", base_demand=1, elevation=10, coordinates=(6, 25)
        )
        duration = 1 * 5 * 60 * 60
        wn.options.time.duration = duration

        # Add fire demand
        node = wn.get_node("new_junction")
        fire_flow_demand = 0.252
        fire_start = 2 * 60 * 60
        fire_end = 4 * 60 * 60
        node.add_fire_fighting_demand(wn, fire_flow_demand, fire_start, fire_end)
        ff_demand = list(
            wntr.metrics.hydraulic.expected_demand(wn)["new_junction"].values
        )
        expected_ff_demand = [1, 1, 1.252, 1.252, 1, 1]
        self.assertListEqual(ff_demand, expected_ff_demand)

        # Remove fire demand
        node.remove_fire_fighting_demand(wn)
        self.assertTrue(
            not ("Fire_Flow" in node.demand_timeseries_list.category_list())
        )


@pytest.mark.parametrize("value", [1.0, 1, "400", 2.71828])
def test_positive_non_zero_float(value):
    assert _check_positive_non_zero_float(value, "test") == float(value)


@pytest.mark.parametrize("value", [0, "0", 0.0, "not_a_number", -1, -2.71828, [-1, 2]])
def test_positive_non_zero_float_bad_value(value):
    with pytest.raises(ValueError, match="test"):
        _check_positive_non_zero_float(value, "test")


@pytest.mark.parametrize("value", [0.0, 0, "0", 1.0, 1, "500", 0, 2.71828])
def test_positive_or_zero_float(value):
    assert _check_positive_or_zero_float(value, "test") == float(value)


@pytest.mark.parametrize("value", ["not_a_number", -1, -2.71828, [-1, 2]])
def test_positive_or_zero_float_bad_value(value):
    with pytest.raises(ValueError, match="test"):
        _check_positive_or_zero_float(value, "test")


@pytest.mark.parametrize("value", [1.0, 1, "400", 2.71828, 0.0, 0, -1, -2.71828])
def test_float_or_none_with_float(value):
    assert _check_float_or_none(value, "test") == float(value)


def test_float_or_none_with_none():
    assert _check_float_or_none(None, "test") is None


@pytest.mark.parametrize("value", ["not_a_number", [-1, 2]])
def test_float_or_none_bad_value(value):
    with pytest.raises(ValueError, match="test"):
        _check_float_or_none(value, "test")


def test_pump_power_on_add_float():
    wn = wntr.network.WaterNetworkModel()
    wn.add_junction("1")
    wn.add_junction("2")

    wn.add_pump("pump1", "1", "2", pump_type="POWER", pump_parameter=1000.0)

    assert wn.links["pump1"].power == 1000.0


def test_pump_power_on_add_int():
    wn = wntr.network.WaterNetworkModel()
    wn.add_junction("1")
    wn.add_junction("2")

    wn.add_pump("pump1", "1", "2", pump_type="POWER", pump_parameter=1000)

    assert wn.links["pump1"].power == 1000.0


@pytest.fixture
def power_pump():
    wn = wntr.network.WaterNetworkModel()
    wn.add_junction("1")
    wn.add_junction("2")
    wn.add_pump("pump1", "1", "2", pump_type="POWER")
    return wn.links["pump1"]


def test_pump_power_set_property_float(power_pump):
    power_pump.power = 500.0

    assert power_pump.power == 500.0


def test_pump_power_set_property_int(power_pump):
    power_pump.power = 500

    assert power_pump.power == 500.0


def test_pump_power_set_property_string(power_pump):
    power_pump.power = "500"

    assert power_pump.power == 500.0


def test_pump_power_set_property_negative(power_pump):
    with pytest.raises(ValueError):
        power_pump.power = -500


def test_pump_power_set_property_zero(power_pump):
    with pytest.raises(ValueError):
        power_pump.power = 0


def test_pump_power_set_property_invalid_string(power_pump):
    with pytest.raises(ValueError):
        power_pump.power = "invalid_string"


if __name__ == "__main__":
    unittest.main()



import unittest
import warnings
from os.path import abspath, dirname, join

import numpy as np
import pandas as pd
import wntr

testdir = dirname(abspath(str(__file__)))
test_network_dir = join(testdir, "networks_for_testing")
test_data_dir = join(testdir, "data_for_testing")
ex_datadir = join(testdir, "..", "..", "examples", "networks")


class Test(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        pass

    @classmethod
    def tearDownClass(self):
        pass

    def test_ReactionVariable_reserved_name_exception(self):
        self.assertRaises(ValueError, wntr.msx.Species,  "D", 'bulk', "mg")
        self.assertRaises(ValueError, wntr.msx.Species,  "Q", 'wall', "mg")
        self.assertRaises(ValueError, wntr.msx.Constant, "Re", 0.52)
        self.assertRaises(ValueError, wntr.msx.Parameter, "Re", 0.52)
        self.assertRaises(ValueError, wntr.msx.Term, "Re", 0.52)

    def test_RxnVariable_values(self):
        const1 = wntr.msx.Constant("Kb", 0.482, note="test")
        self.assertEqual(const1.value, 0.482)
        param2 = wntr.msx.Parameter("Kb", 0.482, note="test")
        self.assertEqual(param2.global_value, 0.482)

    def test_RxnVariable_string_functions(self):
        species1 = wntr.msx.Species('Cl', 'BULK', "mg")
        const1 = wntr.msx.Constant("Kb", 0.482)
        param1 = wntr.msx.Parameter("Ka", 0.482, note="foo")
        term1 = wntr.msx.Term("T0", "-3.2 * Kb * Cl^2", note="bar")

        self.assertEqual(str(species1), "Cl")
        self.assertEqual(str(const1), "Kb")
        self.assertEqual(str(param1), "Ka")
        self.assertEqual(str(term1), "T0")

    def test_Species_tolerances(self):
        # """RxnSpecies(*s) tolerance settings"""
        species1 = wntr.msx.Species("Cl", 'bulk', "mg")
        species2 = wntr.msx.Species("Cl", 'wall', "mg", 0.01, 0.0001, note="Testing stuff")
        self.assertIsNone(species1.get_tolerances())
        self.assertIsNotNone(species2.get_tolerances())
        self.assertTupleEqual(species2.get_tolerances(), (0.01, 0.0001))
        self.assertRaises(TypeError, species1.set_tolerances, None, 0.0001)
        self.assertRaises(ValueError, species1.set_tolerances, -0.51, 0.01)
        self.assertRaises(ValueError, species1.set_tolerances, 0.0, 0.0)
        species1.set_tolerances(0.01, 0.0001)
        self.assertEqual(species1.atol, 0.01)
        self.assertEqual(species1.rtol, 0.0001)
        species1.set_tolerances(None, None)
        self.assertIsNone(species1.atol)
        self.assertIsNone(species1.rtol)
        species2.clear_tolerances()
        self.assertIsNone(species1.atol)
        self.assertIsNone(species1.rtol)
        self.assertIsNone(species2.get_tolerances())

    def test_BulkSpecies_creation(self):
        # """BlukSpecies object creation (direct instantiation)"""
        self.assertRaises(TypeError, wntr.msx.Species, "Re", 'bulk')
        self.assertRaises(ValueError, wntr.msx.Species, "Re", 'bulk', "mg")
        # self.assertRaises(TypeError, wntr.msx.Species, "Cl", 1, "mg", 0.01, None)
        # self.assertRaises(TypeError, wntr.msx.Species, "Cl", wntr.msx.base.SpeciesType.BULK, "mg", None, 0.01)
        species = wntr.msx.Species("Cl", 1, "mg")
        self.assertEqual(species.name, "Cl")
        self.assertEqual(species.units, "mg")
        self.assertEqual(species.species_type, wntr.msx.SpeciesType.BULK)
        self.assertIsNone(species.atol)
        self.assertIsNone(species.rtol)
        self.assertIsNone(species.note)
        species = wntr.msx.Species("Cl",'bulk', "mg", 0.01, 0.0001, note="Testing stuff")
        self.assertEqual(species.species_type, wntr.msx.SpeciesType.BULK)
        self.assertEqual(species.name, "Cl")
        self.assertEqual(species.units, "mg")
        self.assertEqual(species.atol, 0.01)
        self.assertEqual(species.rtol, 0.0001)
        self.assertEqual(species.note, "Testing stuff")

    def test_WallSpecies_creation(self):
        # """WallSpecies object creation (direct instantiation)"""
        self.assertRaises(TypeError, wntr.msx.Species, "Re", 'W')
        self.assertRaises(ValueError, wntr.msx.Species, "Re", 'Wall', "mg")
        self.assertRaises(TypeError, wntr.msx.Species, "Cl", 2, "mg", 0.01)
        self.assertRaises(TypeError, wntr.msx.Species, "Cl", 'w', "mg", None, 0.01)
        species = wntr.msx.Species( "Cl", 'w', "mg")
        self.assertEqual(species.name, "Cl")
        self.assertEqual(species.units, "mg")
        self.assertEqual(species.species_type, wntr.msx.SpeciesType.WALL)
        self.assertIsNone(species.atol)
        self.assertIsNone(species.rtol)
        self.assertIsNone(species.note)
        species = wntr.msx.Species( "Cl", 'w', "mg", 0.01, 0.0001, note="Testing stuff")
        self.assertEqual(species.species_type, wntr.msx.SpeciesType.WALL)
        self.assertEqual(species.name, "Cl")
        self.assertEqual(species.units, "mg")
        self.assertEqual(species.atol, 0.01)
        self.assertEqual(species.rtol, 0.0001)
        self.assertEqual(species.note, "Testing stuff")

    def test_Constant_creation(self):
        self.assertRaises(TypeError, wntr.msx.Constant, "Re")
        self.assertRaises(ValueError, wntr.msx.Constant, "Re", 2.48)
        const1 = wntr.msx.Constant("Kb", 0.482, note="test")
        # FIXME: Find a way to suppress warning printing
        # self.assertWarns(RuntimeWarning, wntr.reaction.Constant, 'Kb1', 0.83, pipe_values={'a',1})
        self.assertEqual(const1.name, "Kb")
        self.assertEqual(const1.value, 0.482)
        self.assertEqual(const1.var_type, wntr.msx.VariableType.CONST)
        self.assertEqual(const1.note, "test")

    def test_Parameter_creation(self):
        self.assertRaises(TypeError, wntr.msx.Parameter, "Re")
        self.assertRaises(ValueError, wntr.msx.Parameter, "Re", 2.48)
        param1 = wntr.msx.Parameter("Kb", 0.482, note="test")
        self.assertEqual(param1.name, "Kb")
        self.assertEqual(param1.global_value, 0.482)
        self.assertEqual(param1.var_type, wntr.msx.VariableType.PARAM)
        self.assertEqual(param1.note, "test")
        # test_pipe_dict = {"PIPE": 0.38}
        # test_tank_dict = {"TANK": 222.23}
        # param2 = wntr.msx.Parameter("Kb", 0.482, note="test", pipe_values=test_pipe_dict, tank_values=test_tank_dict)
        # self.assertDictEqual(param2.pipe_values, test_pipe_dict)
        # self.assertDictEqual(param2.tank_values, test_tank_dict)

    def test_RxnTerm_creation(self):
        self.assertRaises(TypeError, wntr.msx.Term, "Re")
        self.assertRaises(ValueError, wntr.msx.Term, "Re", "1.0*Re")
        term1 = wntr.msx.Term("T0", "-3.2 * Kb * Cl^2", note="bar")
        self.assertEqual(term1.name, "T0")
        self.assertEqual(term1.expression, "-3.2 * Kb * Cl^2")
        self.assertEqual(term1.var_type, wntr.msx.VariableType.TERM)
        self.assertEqual(term1.note, "bar")

    def test_Reaction(self):
        equil1 = wntr.msx.Reaction("Cl", wntr.msx.ReactionType.PIPE, 'equil', "-Ka + Kb * Cl + T0")
        self.assertEqual(equil1.species_name, "Cl")
        self.assertEqual(equil1.expression, "-Ka + Kb * Cl + T0")
        self.assertEqual(equil1.expression_type, wntr.msx.ExpressionType.EQUIL)
        rate1 = wntr.msx.Reaction("Cl", wntr.msx.ReactionType.TANK, 'rate', "-Ka + Kb * Cl + T0", note="Foo Bar")
        self.assertEqual(rate1.species_name, "Cl")
        self.assertEqual(rate1.expression, "-Ka + Kb * Cl + T0")
        self.assertEqual(rate1.expression_type, wntr.msx.ExpressionType.RATE)
        self.assertEqual(rate1.note, "Foo Bar")
        formula1 = wntr.msx.Reaction("Cl", wntr.msx.ReactionType.PIPE, 'formula', "-Ka + Kb * Cl + T0")
        self.assertEqual(formula1.species_name, "Cl")
        self.assertEqual(formula1.expression, "-Ka + Kb * Cl + T0")
        self.assertEqual(formula1.expression_type, wntr.msx.ExpressionType.FORMULA)

    def test_MsxModel_creation_specific_everything(self):
        rxn_model1 = wntr.msx.MsxModel()
        bulk1 = wntr.msx.Species("Cl", 'b', "mg")
        wall1 = wntr.msx.Species("ClOH", 'w', "mg", 0.01, 0.0001, note="Testing stuff")
        const1 = wntr.msx.Constant("Kb", 0.482)
        param1 = wntr.msx.Parameter("Ka", 0.482, note="foo")
        term1 = wntr.msx.Term("T0", "-3.2 * Kb * Cl^2", note="bar")
        equil1 = wntr.msx.Reaction(bulk1, wntr.msx.ReactionType.PIPE, 'equil', "-Ka + Kb * Cl + T0")
        rate1 = wntr.msx.Reaction(bulk1, wntr.msx.ReactionType.TANK, 'rate', "-Ka + Kb * Cl + T0", note="Foo Bar")
        formula1 = wntr.msx.Reaction(wall1, wntr.msx.ReactionType.PIPE, 'formula', "-Ka + Kb * Cl + T0")

        bulk2 = rxn_model1.add_species("Cl", 'bulk', "mg")
        wall2 = rxn_model1.add_species("ClOH", 'wall', "mg", 0.01, 0.0001, note="Testing stuff")
        const2 = rxn_model1.add_constant("Kb", 0.482)
        param2 = rxn_model1.add_parameter("Ka", 0.482, note="foo")
        term2 = rxn_model1.add_term("T0", "-3.2 * Kb * Cl^2", note="bar")
        equil2 = rxn_model1.add_reaction("Cl", "pipe", "equil", "-Ka + Kb * Cl + T0")
        rate2 = rxn_model1.add_reaction("Cl", "tank", wntr.msx.ExpressionType.R, "-Ka + Kb * Cl + T0", note="Foo Bar")
        formula2 = rxn_model1.add_reaction("ClOH", "PIPE", "formula", "-Ka + Kb * Cl + T0")
        rxn_model1.add_source("CONCEN", 'Cl', 'Dummy1', 1.23, note="foo")
        rxn_model1.add_source("SETPOINT", 'Cl', 'Dummy2', 2.7)
        source = rxn_model1.get_sources_for_species('Cl')
        self.assertDictEqual(source, {'Dummy1':{'strength':1.23, 'source_type': 'CONCEN', 'pattern': None, 'note': 'foo'}, 
                             'Dummy2':{'strength':2.7, 'source_type': 'SETPOINT', 'pattern': None, 'note': None}})
        rxn_model1.remove_source('Cl','Dummy1')
        source = rxn_model1.get_sources_for_species('Cl')
        self.assertDictEqual(source, {'Dummy2':{'strength':2.7, 'source_type': 'SETPOINT', 'pattern': None, 'note': None}})
        rxn_model1.remove_sources_for_species('Cl')
        source = rxn_model1.get_sources_for_species('Cl')
        self.assertDictEqual(source, {})
        self.assertDictEqual(bulk1.to_dict(), bulk2.to_dict())
        self.assertDictEqual(wall1.to_dict(), wall2.to_dict())
        self.assertDictEqual(const1.to_dict(), const2.to_dict())
        self.assertDictEqual(param1.to_dict(), param2.to_dict()) 
        self.assertDictEqual(term1.to_dict(), term2.to_dict())
        self.assertDictEqual(equil1.to_dict(), equil2.to_dict())
        self.assertDictEqual(rate1.to_dict(), rate2.to_dict())
        self.assertDictEqual(formula1.to_dict(), formula2.to_dict())


if __name__ == "__main__":
    unittest.main(verbosity=2)



# coding: utf-8
"""
The wntr.msx.options module includes options for multi-species water quality
models
"""

import logging
from typing import Dict, List, Literal, Union

from wntr.network.options import _OptionsBase

logger = logging.getLogger(__name__)


class MsxReportOptions(_OptionsBase):
    """
    Report options

    Parameters
    ----------
    report_filename : str
        Filename for the EPANET-MSX report file, by default this will be
        the prefix plus ".rpt".
    species : dict[str, bool]
        Output species concentrations
    species_precision : dict[str, float]
        Output species concentrations with the specified precision
    nodes : None, "ALL", or list
        Output node information. If a list of node names is provided,
        EPANET-MSX only provides report information for those nodes.
    links : None, "ALL", or list
        Output link information. If a list of link names is provided,
        EPANET-MSX only provides report information for those links.
    pagesize : str
        Page size for EPANET-MSX report output

    """

    def __init__(
        self,
        pagesize: list = None,
        report_filename: str = None,
        species: Dict[str, bool] = None,
        species_precision: Dict[str, float] = None,
        nodes: Union[Literal["ALL"], List[str]] = None,
        links: Union[Literal["ALL"], List[str]] = None,
    ):
        self.pagesize = pagesize
        """Pagesize for the report"""
        self.report_filename = report_filename
        """Prefix of the report filename (will add .rpt)"""
        self.species = species if species is not None else dict()
        """Turn individual species outputs on and off, by default no species are output"""
        self.species_precision = species_precision if species_precision is not None else dict()
        """Output precision for the concentration of a specific species"""
        self.nodes = nodes
        """List of nodes to print output for, or 'ALL' for all nodes, by default None"""
        self.links = links
        """List of links to print output for, or 'ALL' for all links, by default None"""

    def __setattr__(self, name, value):
        if name not in ["pagesize", "report_filename", "species", "nodes", "links", "species_precision"]:
            raise AttributeError("%s is not a valid attribute of ReportOptions" % name)
        self.__dict__[name] = value


class MsxSolverOptions(_OptionsBase):
    """
    Solver options

    Parameters
    ----------
    timestep : int >= 1
        Water quality timestep (seconds), by default 60 (one minute).
    area_units : str, optional
        Units of area to use in surface concentration forms, by default
        ``M2``. Valid values are ``FT2``, ``M2``, or ``CM2``.
    rate_units : str, optional
        Time units to use in all rate reactions, by default ``MIN``. Valid
        values are ``SEC``, ``MIN``, ``HR``, or ``DAY``.
    solver : str, optional
        Solver to use, by default ``RK5``. Options are ``RK5`` (5th order
        Runge-Kutta method), ``ROS2`` (2nd order Rosenbrock method), or
        ``EUL`` (Euler method).
    coupling : str, optional
        Use coupling method for solution, by default ``NONE``. Valid
        options are ``FULL`` or ``NONE``.
    atol : float, optional
        Absolute concentration tolerance, by default 0.01 (regardless of
        species concentration units).
    rtol : float, optional
        Relative concentration tolerance, by default 0.001 (±0.1%).
    compiler : str, optional
        Whether to use a compiler, by default ``NONE``. Valid options are
        ``VC``, ``GC``, or ``NONE``
    segments : int, optional
        Maximum number of segments per pipe (MSX 2.0 or newer only), by
        default 5000.
    peclet : int, optional
        Peclet threshold for applying dispersion (MSX 2.0 or newer only),
        by default 1000.
    report : MsxReportOptions or dict
        Options on how to report out results.

    """

    def __init__(
        self,
        timestep: int = 360,
        area_units: str = "M2",
        rate_units: str = "MIN",
        solver: str = "RK5",
        coupling: str = "NONE",
        atol: float = 1.0e-4,
        rtol: float = 1.0e-4,
        compiler: str = "NONE",
        segments: int = 5000,
        peclet: int = 1000,
        # global_initial_quality: Dict[str, float] = None,
        report: MsxReportOptions = None,
    ):
        self.timestep: int = timestep
        """Timestep, in seconds, by default 360"""
        self.area_units: str = area_units
        """Units used to express pipe wall surface area where, by default FT2.
        Valid values are FT2, M2, and CM2."""
        self.rate_units: str = rate_units
        """Units in which all reaction rate terms are expressed, by default HR.
        Valid values are HR, MIN, SEC, and DAY."""
        self.solver: str = solver
        """Solver to use, by default EUL. Valid values are EUL, RK5, and
        ROS2."""
        self.coupling: str = coupling
        """Whether coupling should occur during solving, by default NONE. Valid
        values are NONE and FULL."""
        self.rtol: float = rtol
        """Relative tolerance used during solvers ROS2 and RK5, by default
        0.001 for all species. Can be overridden on a per-species basis."""
        self.atol: float = atol
        """Absolute tolerance used by the solvers, by default 0.01 for all
        species regardless of concentration units. Can be overridden on a
        per-species basis."""
        self.compiler: str = compiler
        """Compiler to use if the equations should be compiled by EPANET-MSX,
        by default NONE. Valid options are VC, GC and NONE."""
        self.segments: int = segments
        """Number of segments per-pipe to use, by default 5000."""
        self.peclet: int = peclet
        """Threshold for applying dispersion, by default 1000."""
        self.report: MsxReportOptions = MsxReportOptions.factory(report)
        """Reporting output options."""

    def __setattr__(self, name, value):
        if name == "report":
            if not isinstance(value, (MsxReportOptions, dict, tuple, list)):
                raise ValueError("report must be a ReportOptions or convertable object")
            value = MsxReportOptions.factory(value)
        elif name in {"timestep"}:
            try:
                value = max(1, int(value))
            except ValueError:
                raise ValueError("%s must be an integer >= 1" % name)
        elif name in ["atol", "rtol"]:
            try:
                value = float(value)
            except ValueError:
                raise ValueError("%s must be a number", name)
        elif name in ["segments", "peclet"]:
            try:
                value = int(value)
            except ValueError:
                raise ValueError("%s must be a number", name)
        elif name not in ["area_units", "rate_units", "solver", "coupling", "compiler"]:
            raise ValueError("%s is not a valid member of MsxSolverOptions")
        self.__dict__[name] = value

    def to_dict(self):
        """Dictionary representation of the options"""
        return dict(self)



import sys
import unittest
import warnings
import os
from os.path import abspath, dirname, isfile, join

import numpy as np
import pandas as pd
import networkx as nx
import wntr
from pandas.testing import assert_frame_equal, assert_series_equal

try:
    from shapely.geometry import LineString, Point, Polygon, shape
    has_shapely = True
except ModuleNotFoundError:
    has_shapely = False

try:
    import geopandas as gpd
    has_geopandas = True
except ModuleNotFoundError:
    gpd = None
    has_geopandas = False
    
try:
    import rasterio
    has_rasterio = True
except ModuleNotFoundError:
    rasterio = None
    has_rasterio = False
    
testdir = dirname(abspath(str(__file__)))
datadir = join(testdir, "networks_for_testing")
ex_datadir = join(testdir, "..", "..", "examples", "networks")


@unittest.skipIf(not has_geopandas,
                 "Cannot test GIS capabilities: geopandas is missing")
class TestGIS(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        import wntr

        self.wntr = wntr

        inp_file = join(ex_datadir, "Net1.inp")
        self.wn = self.wntr.network.WaterNetworkModel(inp_file)
        sim = wntr.sim.EpanetSimulator(self.wn)
        self.results = sim.run_sim()
        self.gis_data = self.wn.to_gis()
        
        vertex_inp_file = join(datadir, "io.inp")
        self.vertex_wn = self.wntr.network.WaterNetworkModel(vertex_inp_file)
        self.vertex_gis_data = self.vertex_wn.to_gis()
        
        polygon_pts = [[(25,80), (65,80), (65,60), (25,60)],
                       [(25,60), (80,60), (80,30), (25,30)],
                       [(40,50), (60,65), (60,15), (40,15)]]
        polygon_data = []
        for i, pts in enumerate(polygon_pts):
            geometry = Polygon(pts)
            polygon_data.append({'name': str(i+1), 
                                 'value': (i+1)*10,
                                 'geometry': geometry.convex_hull})
            
        df = pd.DataFrame(polygon_data)
        df.set_index("name", inplace=True)
        self.polygons = gpd.GeoDataFrame(df, crs=None)

        points = [(52,71), (75,40), (27,37)]
        point_data = []
        for i, pts in enumerate(points):
            geometry = Point(pts)
            point_data.append({'geometry': geometry})
            
        df = pd.DataFrame(point_data)
        self.points = gpd.GeoDataFrame(df, crs=None)

    @classmethod
    def tearDownClass(self):
        pass
    
    def test_gis_index(self):
        # Tests that WN can be made using dataframes with customized index names
        wn_gis = self.wn.to_gis()
        
        # check that index name of geodataframes is "name"
        assert wn_gis.junctions.index.name == "name"
        assert wn_gis.tanks.index.name == "name"
        assert wn_gis.reservoirs.index.name == "name"
        assert wn_gis.pipes.index.name == "name"
        assert wn_gis.pumps.index.name == "name"
        
        # check that index names can be changed and still be read back into a wn
        wn_gis.junctions.index.name = "my_index"
        wn_gis.pipes.index.name = "my_index"
        wn2 = wntr.network.from_gis(wn_gis)
        
        assert self.wn.pipe_name_list == wn2.pipe_name_list
        assert self.wn.junction_name_list == wn2.junction_name_list
        
        # test snap and intersect functionality with alternate index names
        result = wntr.gis.snap(self.points, wn_gis.junctions, tolerance=5.0)
        assert len(result) > 0
        result = wntr.gis.snap(wn_gis.junctions, self.points, tolerance=5.0)
        assert len(result) > 0
        result = wntr.gis.intersect(wn_gis.junctions, self.polygons)
        assert len(result) > 0
        result = wntr.gis.intersect(self.polygons, wn_gis.pipes)
        assert len(result) > 0
        
        # check that custom index name persists after running snap/intersect
        assert wn_gis.junctions.index.name == "my_index"
        assert wn_gis.pipes.index.name == "my_index"

    def test_wn_to_gis(self):
        # Check type
        isinstance(self.gis_data.junctions, gpd.GeoDataFrame)
        isinstance(self.gis_data.tanks, gpd.GeoDataFrame)
        isinstance(self.gis_data.reservoirs, gpd.GeoDataFrame)
        isinstance(self.gis_data.pipes, gpd.GeoDataFrame)
        isinstance(self.gis_data.pumps, gpd.GeoDataFrame)
        isinstance(self.gis_data.valves, gpd.GeoDataFrame)
        
        # Check size
        assert self.gis_data.junctions.shape[0] == self.wn.num_junctions
        assert self.gis_data.tanks.shape[0] == self.wn.num_tanks
        assert self.gis_data.reservoirs.shape[0] == self.wn.num_reservoirs
        assert self.gis_data.pipes.shape[0] == self.wn.num_pipes
        assert self.gis_data.pumps.shape[0] == self.wn.num_pumps
        #assert self.gis_data.valves.shape[0] == self.wn.num_valves
        
        # Check minimal set of attributes
        assert set(['elevation', 'geometry']).issubset(self.gis_data.junctions.columns)
        assert set(['elevation', 'geometry']).issubset(self.gis_data.tanks.columns)
        assert set(['geometry']).issubset(self.gis_data.reservoirs.columns)
        assert set(['start_node_name', 'end_node_name', 'geometry']).issubset(self.gis_data.pipes.columns)
        assert set(['start_node_name', 'end_node_name', 'geometry']).issubset(self.gis_data.pumps.columns)
        #assert set(['start_node_name', 'end_node_name', 'geometry']).issubset(self.gis_data.valves.columns) # Net1 has no valves

        #check base_demand and demand_pattern attrivutes
        assert set(['base_demand','demand_pattern']).issubset(self.gis_data.junctions.columns)
        
    def test_gis_to_wn(self):
        
        wn2 = wntr.network.io.from_gis(self.gis_data)
        G1 = self.wn.to_graph()
        G2 = wn2.to_graph()
        
        assert nx.is_isomorphic(G1, G2)
        
        # test vertices
        vertex_wn2 = wntr.network.io.from_gis(self.vertex_gis_data)
        for name, link in vertex_wn2.links():
            assert link.vertices == self.vertex_wn.get_link(name).vertices
                         
    def test_intersect_points_with_polygons(self):
        
        stats = wntr.gis.intersect(self.gis_data.junctions, self.polygons, 'value')
        
        # Junction 22 intersects poly2 val=20, intersects poly3 val=30
        # weighted mean = (1*20+0.5*30)/2 = 17.5
        expected = pd.Series({'intersections': ['2','3'], 'values': [20,30]})
        expected['n'] = len(expected['values'])
        expected['sum'] = float(sum(expected['values']))
        expected['min'] = float(min(expected['values']))
        expected['max'] = float(max(expected['values']))
        expected['mean'] = expected['sum']/expected['n']
        expected = expected.reindex(stats.columns)
        
        assert_series_equal(stats.loc['22',:], expected, check_dtype=False, check_names=False)
        
        # Junction 31: no intersections
        expected = pd.Series({'intersections': [], 'values': [], 'n': 0, 
                              'sum': np.nan, 'min': np.nan, 'max': np.nan,
                              'mean': np.nan, })
        
        assert_series_equal(stats.loc['31',:], expected, check_dtype=False, check_names=False)
        
    def test_intersect_lines_with_polygons(self):
        
        bv = 0
        stats = wntr.gis.intersect(self.gis_data.pipes, self.polygons, 'value', True, bv)

        ax = self.polygons.plot(column='value', alpha=0.5)
        ax = wntr.graphics.plot_network(self.wn, ax=ax)
        
        # Pipe 22 intersects poly2 100%, val=20, intersects poly3 50%, val=30
        expected_weighted_mean = (20*1+30*0.5)/1.5
        expected = pd.Series({'intersections': ['2','3'], 'values': [20,30], 'weighted_mean': expected_weighted_mean})
        expected['n'] = len(expected['values'])
        expected['sum'] = float(sum(expected['values']))
        expected['min'] = float(min(expected['values']))
        expected['max'] = float(max(expected['values']))
        expected['mean'] = expected['sum']/expected['n']
        expected = expected.reindex(stats.columns)
        
        assert_series_equal(stats.loc['22',:], expected, check_dtype=False, check_names=False)
        
        # Pipe 31: no intersections
        expected = pd.Series({'intersections': ['BACKGROUND'], 'values': [bv],
                              'n': 1, 'sum': bv, 'min': bv, 'max': bv,
                              'mean': bv, 'weighted_mean': bv})
        
        assert_series_equal(stats.loc['31',:], expected, check_dtype=False, check_names=False)
        
        # Pipe 122
        self.assertEqual(stats.loc['122','intersections'], ['BACKGROUND', '2', '3'])
        # total length = 30
        expected_weighted_mean = (bv*(5/30) + 30*(25/30) + 20*(10/30))/(40/30)
        self.assertAlmostEqual(stats.loc['122','weighted_mean'], expected_weighted_mean, 2)
        
    
    def test_intersect_polygons_with_lines(self):
        
        stats = wntr.gis.intersect(self.polygons, self.gis_data.pipes)
        
        expected = pd.DataFrame([{'intersections': ['10', '11', '110', '111', '112', '12'], 'n': 6},
                                 {'intersections': ['111', '112', '113', '121', '122', '21', '22'], 'n': 7},
                                 {'intersections': ['112', '122', '21', '22'], 'n': 4}])
        expected.index=['1', '2', '3']
    
        assert_frame_equal(stats, expected, check_dtype=False)
        
    def test_intersect_polygons_with_lines_zero_length(self):
        
        wn = wntr.morph.break_pipe(self.wn, '22', '22_A', '22_A','22_B', split_at_point=0.5)
        wn.add_pipe('22_0', '22_A', '22_B', length=0)
        
        gis_data = wn.to_gis()
        assert gis_data.pipes.length['22_0'] == 0
        
        # No value
        stats = wntr.gis.intersect(gis_data.pipes, self.polygons)
        assert stats.shape == (14,2)
        
        # With value
        stats = wntr.gis.intersect(gis_data.pipes, self.polygons, 'value')
        assert stats.shape == (14,8)
        
        assert stats.loc['22_0', 'weighted_mean'] == stats.loc['22_0', 'mean'] # zero length pipe
        assert stats.loc['22_A', 'weighted_mean'] == 20 # overlaps with 20 and 30, but zero length with 30
        assert stats.loc['22', 'weighted_mean'] == 25 # overlaps with 20 and 30 across the entire pipe
        
        assert (stats.loc[stats['n']>0, 'weighted_mean'] >= stats.loc[stats['n']>0, 'min']).all()
        assert (stats.loc[stats['n']>0, 'weighted_mean'] <= stats.loc[stats['n']>0, 'max']).all()
        
    def test_set_crs_to_crs(self):
        # Test uses transformation from https://epsg.io/
        # https://epsg.io/transform#s_srs=4326&t_srs=3857&x=20.0000000&y=70.0000000
        
        gis_data = self.wn.to_gis()
        
        gis_data.set_crs('EPSG:4326')  # EPSG:4326 WGS 84
        x0 = gis_data.junctions.loc['10','geometry'].x
        y0 = gis_data.junctions.loc['10','geometry'].y
        self.assertEqual(gis_data.junctions.crs, 'EPSG:4326')
        self.assertEqual(x0, 20)
        self.assertEqual(y0, 70)
        
        gis_data.to_crs('EPSG:3857')  # EPSG:3857 WGS 84 / Pseudo-Mercator
        x1 = gis_data.junctions.loc['10','geometry'].x
        y1 = gis_data.junctions.loc['10','geometry'].y
        self.assertEqual(gis_data.junctions.crs, 'EPSG:3857')
        self.assertAlmostEqual(x1, 2226389.8158654715, 6)
        self.assertAlmostEqual(y1, 11068715.659379493, 6)
        
    def test_add_attributes_and_write(self):
        
        gis_data = self.wn.to_gis()
        
        gis_data.add_node_attributes(self.results.node['pressure'].loc[3600,:], 'Pressure_1hr')
        gis_data.add_link_attributes(self.results.link['flowrate'].loc[3600,:], 'Flowrate_1hr')
       
        assert 'Pressure_1hr' in gis_data.junctions.columns
        assert 'Pressure_1hr' in gis_data.tanks.columns
        assert 'Pressure_1hr' in gis_data.reservoirs.columns
        assert 'Flowrate_1hr' in gis_data.pipes.columns
        assert 'Flowrate_1hr' in gis_data.pumps.columns
        assert 'Flowrate_1hr' not in gis_data.valves.columns # Net1 has no valves
    
    def test_write_geojson(self):
        prefix = 'temp_Net1'
        components = ['junctions', 'tanks', 'reservoirs', 'pipes', 'pumps', 'valves']
        for component in components:
            filename = abspath(join(testdir, prefix+'_'+component+'.geojson'))
            if isfile(filename):
                os.remove(filename)
            
        self.gis_data.write_geojson(abspath(join(testdir, prefix)))

        for component in components:
            if component == 'valves':
                continue # Net1 has no valves
            # check file exists
            filename = abspath(join(testdir, prefix+'_'+component+'.geojson'))
            self.assertTrue(isfile(filename))
            
            # check for "name" column
            gdf = gpd.read_file(filename)
            assert "name" in gdf.columns

    def test_snap_points_to_points(self):
        
        snapped_points = wntr.gis.snap(self.points, self.gis_data.junctions, tolerance=5.0)        
        # distance = np.sqrt(2)*2, 5, np.sqrt(2)*3
        expected = pd.DataFrame([{'node': '12', 'snap_distance': 2.23607, 'geometry': Point([50.0,70.0])},
                                 {'node': '23', 'snap_distance': 5.0,      'geometry': Point([70.0,40.0])},
                                 {'node': '21', 'snap_distance': 4.242641, 'geometry': Point([30.0,40.0])}])
        
        assert_frame_equal(pd.DataFrame(snapped_points), expected, check_dtype=False)
        

    def test_snap_points_to_lines(self):
        
        snapped_points = wntr.gis.snap(self.points, self.gis_data.pipes, tolerance=5.0)
        
        # distance = 1,5,3
        expected = pd.DataFrame([{'link': '12', 'node': '12', 'snap_distance': 1, 'line_position': 0.1, 'geometry': Point([52.0,70.0])},
                                 {'link': '113', 'node': '23', 'snap_distance': 5.0, 'line_position': 1.0, 'geometry': Point([70.0,40.0])},
                                 {'link': '121', 'node': '21', 'snap_distance': 3.0, 'line_position': 0.1, 'geometry': Point([30.0,37.0])}])
        
        assert_frame_equal(pd.DataFrame(snapped_points), expected, check_dtype=False)


@unittest.skipIf(not has_geopandas,
                 "Cannot test GIS capabilities: geopandas is missing")
class TestConnectLines(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        
        np.random.seed(456)
        
        inp_file = join(ex_datadir, "Net1.inp")
        wn = wntr.network.WaterNetworkModel(inp_file)
        wn_gis = wn.to_gis(crs="EPSG:2236") # ft, https://epsg.io/2236, NAD83 / Florida East (ftUS)
        original_pipes = wn_gis.pipes
        
        # Create imperfect pipe data
        disconnected_pipes = original_pipes.copy()
        for i, line in disconnected_pipes.iterrows():
            angle = np.random.uniform(-5,5,1)
            geom = gpd.GeoSeries(line['geometry'])
            geom = geom.rotate(angle)
            geom = geom.scale(0.9, 0.9)
            disconnected_pipes.loc[i,'geometry'] = geom[0]
        
        self.original_pipes = wn_gis.pipes
        self.disconnected_pipes = disconnected_pipes
 
    @classmethod
    def tearDownClass(self):
        pass
    
    def test_connect_lines(self):
        distance_threshold = 5 # same units as crs (ft)
        pipes, junctions = wntr.gis.connect_lines(self.disconnected_pipes, 
                                                  distance_threshold)
        
        gis_data = wntr.gis.WaterNetworkGIS({"junctions": junctions,
                                             "pipes": pipes})
        wn = wntr.network.from_gis(gis_data)
        G = wn.to_graph()
        
        uG = G.to_undirected()
        assert nx.is_connected(uG)
        assert nx.number_connected_components(uG) == 1


@unittest.skipIf(not has_rasterio,
                 "Cannot test raster capabilities: rasterio is missing")
class TestRaster(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        # use net1 junctions as example points
        inp_file = join(ex_datadir, "Net1.inp")
        wn = wntr.network.WaterNetworkModel(inp_file)
        wn_gis = wn.to_gis(crs="EPSG:4326")
        points = pd.concat((wn_gis.junctions, wn_gis.tanks))
        self.points = points
        
        min_lon, min_lat, max_lon, max_lat = self.points.total_bounds

        resolution = 1.0
        
        # adjust to include boundary
        max_lon += resolution
        min_lat -= resolution
                
        lon_values = np.arange(min_lon, max_lon, resolution)
        lat_values = np.arange(max_lat, min_lat, -resolution)  # Decreasing order for latitudes
        raster_data = np.outer(lat_values,lon_values) # value is product of coordinate

        transform = rasterio.transform.from_origin(min_lon, max_lat, resolution, resolution)
        with rasterio.open(
            "test_raster.tif", "w", driver="GTiff", height=raster_data.shape[0], width=raster_data.shape[1], 
            count=1, dtype=raster_data.dtype, crs="EPSG:4326", transform=transform) as file:
            file.write(raster_data, 1) 
        
    @classmethod
    def tearDownClass(self):
        pass
    
    def test_sample_raster(self):
        raster_values = wntr.gis.sample_raster(self.points, "test_raster.tif")
        assert (raster_values.index == self.points.index).all()
        
        # values should be product of coordinates
        expected_values = self.points.apply(lambda row: row.geometry.x * row.geometry.y, axis=1)
        assert np.isclose(raster_values.values, expected_values, atol=1e-5).all()


if __name__ == "__main__":
    unittest.main()


"""
The wntr.network.base module includes base classes for network elements and 
the network model.
"""
import logging
import six
from six import string_types
import types
from wntr.utils.ordered_set import OrderedSet

import enum
import sys
from collections.abc import MutableSequence, MutableMapping
from collections import OrderedDict
from wntr.utils.ordered_set import OrderedSet

import abc

logger = logging.getLogger(__name__)


class AbstractModel(object):
    """
    Base class for water network models.
    """
    pass


class Subject(object):
    """
    Base class for the subject in an observer design pattern.
    """
    def __init__(self):
        self._observers = OrderedSet()

    def subscribe(self, observer):
        self._observers.add(observer)

    def unsubscribe(self, observer):
        self._observers.remove(observer)

    def notify(self):
        for o in self._observers:
            o.update(self)


class Observer(six.with_metaclass(abc.ABCMeta, object)):
    """
    Base class for the observer in an observer design pattern.
    """
    @abc.abstractmethod
    def update(self, subject):
        pass


class Node(six.with_metaclass(abc.ABCMeta, object)):
    """Base class for nodes.
    
    For details about the different subclasses, see one of the following:
    :class:`~wntr.network.elements.Junction`, 
    :class:`~wntr.network.elements.Tank`, and
    :class:`~wntr.network.elements.Reservoir`


    .. rubric:: Constructor
    
    This is an abstract class and should not be instantiated directly.

    Parameters
    -----------
    wn : :class:`~wntr.network.model.WaterNetworkModel`
        WaterNetworkModel object
    name : string
        Name of the node (must be unique among nodes of all types)
    

    .. rubric:: Attributes

    .. autosummary::

        name
        node_type
        coordinates
        initial_quality
        tag
    

    .. rubric:: Read-only simulation results

    The following attributes are read-only. The values are the final calculated
    value from a simulation.

    .. autosummary::

        head
        demand
        leak_demand
        leak_status
        leak_area
        leak_discharge_coeff

    """
    def __init__(self, wn, name):
        self._name = name
        self._head = None
        self._demand = None
        self._pressure = None
        self._quality = None
        self._leak_demand = None
        self._initial_quality = None
        self._tag = None
        self._leak = False
        self._leak_status = False
        self._leak_area = 0.0
        self._leak_discharge_coeff = 0.0
        self._options = wn._options
        self._node_reg = wn._node_reg
        self._link_reg = wn._link_reg
        self._controls = wn._controls
        self._pattern_reg = wn._pattern_reg
        self._curve_reg = wn._curve_reg
        self._coordinates = [0,0]
        self._source = None
        self._is_isolated = False

    def _compare(self, other):
        """
        Comparison function

        Parameters
        ----------
        other : Node
            object to compare with

        Returns
        -------
        bool
            is these the same items
        """        
        if not type(self) == type(other):
            return False
        if self.name == other.name and \
           self.initial_quality == other.initial_quality and \
           self.tag == other.tag:
               return True
        return False

    def __str__(self):
        return self._name

    def __repr__(self):
        return "<Node '{}'>".format(self._name)

    @property
    def head(self):
        """float: (read-only) the current simulation head at the node (total head)"""
        return self._head
    # @head.setter
    # def head(self, value):
    #     self._head = value

    @property
    def demand(self):
        """float: (read-only) the current simulation demand at the node (actual demand)"""
        return self._demand
    # @demand.setter
    # def demand(self, value):
    #     self._demand = value

    @property
    def pressure(self):
        """float : (read-only) the current simulation pressure at the node"""
        return self._pressure

    @property
    def quality(self):
        """float : (read-only) the current simulation quality at the node"""
        return self._quality

    @property
    def leak_demand(self):
        """float: (read-only) the current simulation leak demand at the node"""
        return self._leak_demand
    # @leak_demand.setter
    # def leak_demand(self, value):
    #     self._leak_demand = value

    @property
    def leak_status(self):
        """bool:(read-only) the current simulation leak status at the node"""
        return self._leak_status
    # @leak_status.setter
    # def leak_status(self, value):
    #     self._leak_status = value
    
    @property
    def leak(self):
        """float: (read-only) the current simulation leak area at the node"""
        return self._leak

    @property
    def leak_area(self):
        """float: (read-only) the current simulation leak area at the node"""
        return self._leak_area
    # @leak_area.setter
    # def leak_area(self, value):
    #     self._leak_area = value

    @property
    def leak_discharge_coeff(self):
        """float: (read-only) the current simulation leak discharge coefficient"""
        return self._leak_discharge_coeff
    # @leak_discharge_coeff.setter
    # def leak_discharge_coeff(self, value):
    #     self._leak_discharge_coeff = value

    @property
    def node_type(self):
        """str: The node type (read only)"""
        return 'Node'
    
    @property
    def name(self):
        """str: The name of the node (read only)"""
        return self._name
    
    @property
    def tag(self):
        """str: A tag or label for the node"""
        return self._tag
    @tag.setter
    def tag(self, tag):
        self._tag = tag

    @property
    def initial_quality(self):
        """float or dict: Initial quality (concentration) at the node, or
        a dict of species-->quality for multi-species quality"""
        if not self._initial_quality:
            return 0.0
        return self._initial_quality
    @initial_quality.setter
    def initial_quality(self, value):
        if value and not isinstance(value, (list, float, int, dict)):
            raise ValueError('Initial quality must be a float or a list')
        self._initial_quality = value

    @property
    def coordinates(self):
        """tuple: Node coordinates, (x,y)"""
        return self._coordinates
    @coordinates.setter
    def coordinates(self, coordinates):
        if isinstance(coordinates, (list, tuple)) and len(coordinates) == 2:
            self._coordinates = tuple(coordinates)
        else:
            raise ValueError('coordinates must be a 2-tuple or len-2 list')

    def to_dict(self):
        """Dictionary representation of the node"""
        d = {}
        d['name'] = self.name
        d['node_type'] = self.node_type
        for k in dir(self):
            if not k.startswith('_') and \
              k not in ['demand', 'head', 'leak_demand', 'leak_status', 
                        'level', 'pressure', 'quality', 'vol_curve', 'head_timeseries']:
                try:
                    val = getattr(self, k)
                    if not isinstance(val, types.MethodType):
                        if hasattr(val, "to_ref"):
                            d[k] = val.to_ref()
                        elif hasattr(val, "to_list"):
                            d[k] = val.to_list()
                        elif hasattr(val, "to_dict"):
                            d[k] = val.to_dict()
                        elif isinstance(val, (enum.IntEnum, enum.Enum)):
                            d[k] = str(val)
                        else:
                            d[k] = val
                except DeprecationWarning: pass
        return d

    def to_ref(self):
        return self._name


class Link(six.with_metaclass(abc.ABCMeta, object)):
    """Base class for links.

    For details about the different subclasses, see one of the following:
    :class:`~wntr.network.elements.Pipe`, 
    :class:`~wntr.network.elements.Pump`, and
    :class:`~wntr.network.elements.Valve`

    .. rubric:: Constructor
    
    This is an abstract class and should not be instantiated directly.

    Parameters
    ----------
    wn : :class:`~wntr.network.model.WaterNetworkModel`
        WaterNetworkModel object
    link_name : string
        Name of the link
    start_node_name : string
        Name of the start node
    end_node_name : string
        Name of the end node
    

    .. rubric:: Attributes

    .. autosummary::

        name
        link_type
        start_node
        start_node_name
        end_node
        end_node_name
        initial_status
        initial_setting
        initial_quality
        tag
        vertices

    .. rubric:: Read-only simulation results

    The following attributes are read-only. The values are the final calculated
    value from a simulation.

    .. autosummary::

        flow
        headloss
        quality
        status
        setting

    """
    def __init__(self, wn, link_name, start_node_name, end_node_name):
        # Set the registries
        self._options = wn._options
        self._node_reg = wn._node_reg
        self._link_reg = wn._link_reg
        self._controls = wn._controls
        self._pattern_reg = wn._pattern_reg
        self._curve_reg = wn._curve_reg
        # Set the link name
        self._link_name = link_name
        # Set and register the starting node
        self._start_node = self._node_reg[start_node_name]
        self._node_reg.add_usage(start_node_name, (link_name, self.link_type))
        # Set and register the ending node
        self._end_node = self._node_reg[end_node_name]
        self._node_reg.add_usage(end_node_name, (link_name, self.link_type))
        # Set status variables
        self._user_status = LinkStatus.Opened # Control status, can be changed by the user
        self._initial_status = LinkStatus.Opened # Model initial status, can be changed by the user
        self._internal_status = LinkStatus.Active # Intermediate simulation status, read only
        self._status= None # Current simulation status, read only
        # Set setting variables
        self._initial_setting = None # Model initial setting, can be changed by the user
        self._prev_setting = None # Previous simulation setting, read only
        self._setting = None # Current simulation setting, read only
        # Other model state variables
        self._initial_quality = None
        self._flow = None
        self._velocity = None
        self._is_isolated = False
        self._quality = None
        self._headloss = None
        # Other
        self._vertices = []
        self._tag = None
        

    def _compare(self, other):
        """
        Parameters
        ----------
        other: Link

        Returns
        -------
        bool
        """
        if not type(self) == type(other):
            return False
        if self.name != other.name:
            return False
        if self.tag != other.tag:
            return False
        if self.initial_status != other.initial_status:
            return False
        if (self.initial_setting is not None) ^ (other.initial_setting is not None):
                return False 
        elif (self.initial_setting is not None) and (other.initial_setting is not None):
            if abs(self.initial_setting - other.initial_setting) > 1e-9:
                return False
        if self.start_node_name != other.start_node_name:
            return False
        if self.end_node_name != other.end_node_name:
            return False
        return True
    
    def __str__(self):
        return self._link_name

    def __repr__(self):
        return "<Link '{}'>".format(self._link_name)

    @property
    def link_type(self):
        """str: the link type (read only)"""
        return 'Link'

    @property
    def initial_status(self):
        """:class:`~wntr.network.base.LinkStatus`: The initial status (`Opened`, `Closed`, `Active`) of the Link"""
        return self._initial_status
    @initial_status.setter
    def initial_status(self, status):
        if not isinstance(status, LinkStatus):
            if isinstance(status, int): status = LinkStatus(status)
            elif isinstance(status, str): status = LinkStatus[status]
            else: status = LinkStatus(int(status))
        self._initial_status = status
        self._user_status = status
        
    @property
    def initial_setting(self):
        """float: The initial setting for the link (if `Active`)"""
        return self._initial_setting
    @initial_setting.setter
    def initial_setting(self, setting):
        # TODO: typechecking
        self._initial_setting = setting
        self._setting =  setting

    @property
    def start_node(self):
        """:class:`~wntr.network.base.Node`: The start node object."""
        return self._start_node
    @start_node.setter
    def start_node(self, node):
        self._node_reg.remove_usage(self.start_node_name, (self._link_name, self.link_type))
        self._node_reg.add_usage(node.name, (self._link_name, self.link_type))
        self._start_node = self._node_reg[node.name]

    @property
    def end_node(self):
        """:class:`~wntr.network.base.Node`: The end node object."""
        return self._end_node
    @end_node.setter
    def end_node(self, node):
        self._node_reg.remove_usage(self.end_node_name, (self._link_name, self.link_type))
        self._node_reg.add_usage(node.name, (self._link_name, self.link_type))
        self._end_node = self._node_reg[node.name]

    @property
    def start_node_name(self):
        """str: The name of the start node (read only)"""
        return self._start_node.name
    
    @property
    def end_node_name(self):
        """str: The name of the end node (read only)"""
        return self._end_node.name

    @property
    def name(self):
        """str: The link name (read-only)"""
        return self._link_name

    @property
    def flow(self):
        """float: (read-only) current simulated flow through the link"""
        return self._flow
    
    @property
    def velocity(self):
        """float: (read-only) current simulated velocity through the link"""
        return self._velocity
    
    @property
    @abc.abstractmethod
    def status(self):
        """:class:`~wntr.network.base.LinkStatus`: (**abstract**) current status of the link"""
        pass
    @status.setter
    # @abc.abstractmethod
    def status(self, status):
        raise RuntimeError("The status attribute is an output (result) property. Setting status by"
                            " the user has been deprecated to avoid confusion. To change the simulation"
                            " behavior, use initial_status.")
        # self._user_status = status
    
    @property
    def initial_quality(self):
        """float or dict : a dict of species and quality if multispecies is active"""
        if not self._initial_quality:
            return 0.0
        return self._initial_quality
    @initial_quality.setter
    def initial_quality(self, value):
        if value and not isinstance(value, (list, float, int, dict)):
            raise ValueError('Initial quality must be a float or a list')
        self._initial_quality = value

    @property
    def quality(self):
        """float : (read-only) current simulated average link quality"""
        return self._quality

    @property
    def headloss(self):
        """float : (read-only) current simulated headloss"""
        return self._headloss

    @property
    def setting(self):
        """float: (read-only) current simulated setting of the link"""
        return self._setting
    @setting.setter
    def setting(self, setting):
        raise RuntimeError("The setting attribute is an output (result) property. Setting the setting by"
                            " the user has been deprecated to avoid confusion. To change the simulation"
                            " behavior, use initial_setting.")
        # self._setting = setting
    
    @property
    def tag(self):
        """str: A tag or label for this link"""
        return self._tag
    @tag.setter
    def tag(self, tag):
        self._tag = tag
        
    @property
    def vertices(self):
        """A list of interior vertex points for network links, in the direction of start node to end node.
        
        The vertices should be listed as a list of (x,y) tuples/lists when setting.
        """
        return self._vertices
    @vertices.setter
    def vertices(self, points):
        if not isinstance(points, list):
            raise ValueError('vertices must be a list')
        for pt in points:
            if not isinstance(pt, (tuple, list)) or len(pt) != 2:
                raise ValueError('vertex points must be a tuple or list with two values')
        self._vertices = points
    
    def to_dict(self):
        """Dictionary representation of the link"""
        d = {}
        d['name'] = self.name
        d['link_type'] = self.link_type
        d['start_node_name'] = self.start_node_name
        d['end_node_name'] = self.end_node_name
        if hasattr(self, 'pump_type'):
            d['pump_type'] = self.pump_type
        if hasattr(self, 'valve_type'):
            d['valve_type'] = self.valve_type
        for k in dir(self):
            if not k.startswith('_') and k not in [
                'flow', 'cv', 'friction_factor', 'headloss',
                'quality', 'reaction_rate', 'setting', 'status', 'velocity', 'speed_timeseries', 'efficiency'
            ]:
                val = getattr(self, k)
                if not isinstance(val, types.MethodType):
                    if hasattr(val, "to_ref"):
                        if hasattr(self, k+"_name") and getattr(self, k+"_name") is not None:
                            continue
                        d[k] = val.to_ref()
                    elif hasattr(val, "to_list"):
                        d[k] = val.to_list()
                    elif hasattr(val, "to_dict"):
                        d[k] = val.to_dict()
                    elif isinstance(val, (enum.IntEnum, enum.Enum)):
                        d[k] = str(val)
                    else:
                        d[k] = val
        return d

    def to_ref(self):
        return self._name


class Registry(MutableMapping):
    """Base class for registries.
    
    Parameters
    ----------
    wn : :class:`~wntr.network.model.WaterNetworkModel`
        WaterNetworkModel object
    
    """
    
    def __init__(self, wn):
        if not isinstance(wn, AbstractModel):
            raise ValueError('Registry must be initialized with a model')
#        self._m = model
        self._data = OrderedDict()
        self._usage = OrderedDict()

    def _finalize_(self, wn):
        self._options = wn._options
        self._pattern_reg = wn._pattern_reg
        self._curve_reg = wn._curve_reg
        self._node_reg = wn._node_reg
        self._link_reg = wn._link_reg
        self._controls = wn._controls
        self._sources = wn._sources
    
    def __getitem__(self, key):
        if not key:
            return None
        try:
            return self._data[key]
        except KeyError:
            try:
                return self._data[key.name]
            except:
                return self._data[str(key)]
            

    def __setitem__(self, key, value):
        if not isinstance(key, string_types):
            raise ValueError('Registry keys must be strings')
        self._data[key] = value
    
    def __delitem__(self, key):
        try:
            if self._usage and key in self._usage and len(self._usage[key]) > 0:
                raise RuntimeError('cannot remove %s %s, still used by %s', 
                                   self.__class__.__name__,
                                   key,
                                   self._usage[key])
            elif key in self._usage:
                self._usage.pop(key)
            return self._data.pop(key)
        except KeyError:
            # Do not raise an exception if there is no key of that name
            return
    
    def __iter__(self):
        return self._data.__iter__()
    
    def __len__(self):
        return len(self._data)

    def __call__(self):
        for key, value in self._data.items():
            yield key, value

    def usage(self):
        """Generator to get the usage for all objects in the registry
        
        Yields
        ------
        key : str
            The name of the object in the registry
        value : tuple of (str, str)
            Tuple of (name, typestr) of the external items using the object
        
        """
        for k, v in self._usage.items():
            yield k, v
        
    def get_usage(self, key):
        """Get a set of items using an object by key.
        
        Returns
        -------
        set of 2-tuples
            Set of (name, typestr) of the external object using the item
            
        """
        try:
            return self._usage[key]
        except KeyError:
            try:
                return self._usage[str(key)]
            except KeyError:
                return None
        return None

    def orphaned(self):
        """Get a list of orphaned usages.
        
        If removed without appropriate checks, it is possible that a some other 
        item will point to an object that has been deleted. (This is why the user
        should always use the "remove_*" methods). This method returns a list of 
        names for objects that are referenced, but no longer exist.
        
        Returns
        -------
        set
            The names of any orphaned items
        
        """
        defined = set(self._data.keys())
        assigned = set(self._usage.keys())
        return assigned.difference(defined)
    
    def unused(self):
        """Get a list of items which are unused by other objects in the model.
        
        In most cases, this method will give little information. For nodes, however,
        this method could be important to identify a node which has become completely 
        disconnected from the network. For patterns or curves, it may be used to find
        extra patterns or curves that are no longer necessary (or which the user 
        forgot ot assign). It is not terribly useful for links.
        
        Returns
        -------
        set
            The names of any unused objects in the registry
            
        """
        defined = set(self._data.keys())
        assigned = set(self._usage.keys())
        return defined.difference(assigned)

    def clear_usage(self, key):
        """if key in usage, clear usage[key]"""
        if not key:
            return
        self._usage[key].clear()
    
    def add_usage(self, key, *args):
        """add args to usage[key]"""
        if not key:
            return
        if not (key in self._usage):
            self._usage[key] = OrderedSet()
        for arg in args:
            self._usage[key].add(arg)

    def remove_usage(self, key, *args):
        """remove args from usage[key]"""
        if not key:
            return
        for arg in args:
            self._usage[key].discard(arg)
        if len(self._usage[key]) < 1:
            self._usage.pop(key)

    def to_dict(self):
        """Dictionary representation of the registry"""
        d = dict()
        for k, v in self._data.items():
            d[k] = v.to_dict()
        return d
    
    def to_list(self):
        """List representation of the registry"""
        l = list()
        for k, v in self._data.items():
            l.append(v.to_dict())
        return l


class NodeType(enum.IntEnum):
    """
    Enum class for node types.

    .. rubric:: Enum Members

    .. autosummary::

        Junction
        Reservoir
        Tank


    """
    Junction = 0  #: node is a junction
    Reservoir = 1  #: node is a reservoir
    Tank = 2  #: node is a tank

    def __init__(self, val):
        mmap = getattr(self, '_member_map_')
        if self.name != str(self.name).upper():
            mmap[str(self.name).upper()] = self
        if self.name != str(self.name).lower():
            mmap[str(self.name).lower()] = self

    def __str__(self):
        return self.name

    def __eq__(self, other):
        return int(self) == int(other) and (isinstance(other, int) or \
               self.__class__.__name__ == other.__class__.__name__)


class LinkType(enum.IntEnum):
    """
    Enum class for link types.

    .. rubric:: Enum Members

    .. autosummary::

        CV
        Pipe
        Pump
        PRV
        PSV
        PBV
        FCV
        TCV
        GPV
        Valve


    """
    CV = 0  #: pipe with a check valve
    Pipe = 1  #: pipe with no check valve
    Pump = 2  #: a pump of any type 
    PRV = 3  #: a pressure reducing valve
    PSV = 4  #: a pressure sustaining valve
    PBV = 5  #: a pressure breaker valve
    FCV = 6  #: a flow control valve
    TCV = 7  #: a throttle control valve
    GPV = 8  #: a general purpose valve
    Valve = 9  #: a valve of any type

    def __init__(self, val):
        mmap = getattr(self, '_member_map_')
        if self.name != str(self.name).upper():
            mmap[str(self.name).upper()] = self
        if self.name != str(self.name).lower():
            mmap[str(self.name).lower()] = self

    def __str__(self):
        return self.name

    def __eq__(self, other):
        return int(self) == int(other) and (isinstance(other, int) or \
               self.__class__.__name__ == other.__class__.__name__)


class LinkStatus(enum.IntEnum):
    """
    Enum class for link statuses.

    .. warning::
        This is NOT the class for determining output status from an EPANET **binary** file.
        The class for output status is wntr.epanet.util.LinkTankStatus.

    .. rubric:: Enum Members

    .. autosummary::

        Closed
        Opened
        Active
        CV
        Open


    """
    Closed = 0  #: pipe/valve/pump is closed
    Open = 1  #: alias for `Opened`
    Opened = 1  #: pipe/valve/pump is open
    Active = 2  #: valve is partially open or pump has a specific setting
    CV = 3  #: pipe has a check valve

    def __init__(self, val):
        mmap = getattr(self, '_member_map_')
        if self.name != str(self.name).upper():
            mmap[str(self.name).upper()] = self
        if self.name != str(self.name).lower():
            mmap[str(self.name).lower()] = self

    def __str__(self):
        return self.name

    def __eq__(self, other):
        return int(self) == int(other) and (isinstance(other, int) or
                                            self.__class__.__name__ == other.__class__.__name__)



import unittest
import os
import shutil
from os.path import abspath, dirname, join
from wntr.network import WaterNetworkModel
from wntr.library import ModelLibrary, model_library

testdir = dirname(abspath(str(__file__)))
test_datadir = join(testdir, "networks_for_testing")
ex_datadir = join(testdir, "..", "..", "examples", "networks")

class TestModelLibrary(unittest.TestCase):
    def test_model_name_list(self):
        self.assertIn("Net3", model_library.model_name_list)
        self.assertIn("ky4", model_library.model_name_list)
    
    def test_get_model(self):
        model = model_library.get_model("ky4")
        self.assertIsInstance(model, WaterNetworkModel)
        
    def test_model_construction(self):
        # Test construction from name
        model = WaterNetworkModel("ky4")
        
    def test_add_directory(self):
        # Test adding a directory
        model_library.add_directory(test_datadir)
        self.assertIn("io", model_library.model_name_list)
        
        # Test getting the filepath after adding a directory
        filepath = model_library.get_filepath("io")
        self.assertEqual(filepath, join(test_datadir, "io.inp"))
        
        # Test construction from file path
        model = WaterNetworkModel("io")
        self.assertIsInstance(model, WaterNetworkModel)


if __name__ == "__main__":
    unittest.main()



# coding: utf-8
"""Decorators for use with enum classes.
"""

from enum import Enum
from typing import Union
import functools

def add_get(cls=None, *, prefix=None, abbrev=False, allow_none=True):
    """Decorator that will add a ``get()`` classmethod to an enum class.

    Parameters
    ----------
    prefix : str, optional
        A prefix to strip off any string values passed in, by default None
    abbrev : bool, optional
        Allow truncating to the first character for checks, by default False
    allow_none : bool, optional
        Allow None to be be passed through without raising error, by default True
    
    Returns
    -------
    class
        the modified class

    Notes
    -----
    The ``get`` method behaves as follows:

    For an integer, the integer value will be used to select the proper member.
    For an :class:`Enum` object, the object's ``name`` will be used, and it will
    be processed as a string. For a string, the method will:
    
    0. if ``allow_none`` is False, then raise a TypeError if the value is None, otherwise
       pass None back to calling function for processing
    1. capitalize the string
    2. remove leading or trailing spaces
    3. convert interior spaces or dashes to underscores
    4. optionally, remove a specified prefix from a string (using ``prefix``, which 
        should have a default assigned by the :func:`wntr.utils.enumtools.add_get` 
        function.)

    It will then try to get the member with the name corresponding to the converted
    string.
    
    5. optionally, if ``abbrev`` is True, then the string will be truncated to the first
        letter, only, after trying to use the full string as passed in. The ``abbrev``
        parameter will have a default value based on how the :func:`~wntr.utils.enumtools.add_get`
        decorator was called on this class.
    
    """
    if prefix is None:
        prefix = ''
    if abbrev is None:
        abbrev = False
    if allow_none is None:
        allow_none = True
    if cls is None:
        return functools.partial(add_get, prefix=prefix, abbrev=abbrev, allow_none=allow_none)

    @functools.wraps(cls)
    def wrap(cls, prefix, abbrev):
        """Perform the decorator action"""

        def get(cls, value: Union[str, int, Enum], prefix='', abbrev=False, allow_none=True):
            """Get the proper enum based on the name or value of the argument.

            See :func:`~wntr.utils.enumtools.add_get` for details on how this function works.
            

            Parameters
            ----------
            value : Union[str, int, Enum]
                the value to be checked, if it is an Enum, then the name will be used
            prefix : str, optional
                a prefix to strip from the beginning of ``value``, default blank or set by decorator
            abbrev : bool, optional
                whether to try a single-letter version of ``value``, default False or set by decorator
            allow_none : bool, optional
                passing None will return None, otherwise will raise TypeError, default True or set by decorator

            Returns
            -------
            Enum
                the enum member that corresponds to the name or value passed in

            Raises
            ------
            TypeError
                if ``value`` is an invalid type
            ValueError
                if ``value`` is invalid

            """
            if value is None and allow_none:
                return None
            elif value is None:
                raise TypeError('A value is mandatory, but got None')
            name = str(value)
            if isinstance(value, cls):
                return value
            elif isinstance(value, int):
                return cls(value)
            elif isinstance(value, str):
                name = value.upper().strip().replace('-', '_').replace(' ', '_')
                if name.startswith(prefix):
                    name = name[len(prefix):]
            elif isinstance(value, Enum):
                name = str(value.name).upper().strip().replace('-', '_').replace(' ', '_')
                if name.startswith(prefix):
                    name = name[len(prefix):]
            else:
                raise TypeError('Invalid type for value: %s'%type(value))
            if abbrev:
                try:
                    return cls[name]
                except KeyError as e:
                    try:
                        return cls[name[0]]
                    except KeyError:
                        raise ValueError(repr(value)) from e
            else:
                try:
                    return cls[name]
                except KeyError as e:
                    raise ValueError(repr(value)) from e
        
        setattr(cls, "get", classmethod(functools.partial(get, prefix=prefix, abbrev=abbrev)))
        return cls

    return wrap(cls, prefix, abbrev)




import unittest
from pandas.testing import assert_series_equal
import os
from os.path import abspath, dirname, join, isfile
import numpy as np
import pandas as pd
import matplotlib.pylab as plt

import wntr
from wntr.library import DemandPatternLibrary

testdir = dirname(abspath(str(__file__)))
test_datadir = join(testdir, "networks_for_testing")
ex_datadir = join(testdir, "..", "..", "examples", "networks")

plt.close('all')

class TestDemandPatternLibrary(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        self.DPL = DemandPatternLibrary()
        
    @classmethod
    def tearDownClass(self):
        pass

    def test_pattern_name_list(self):
        pattern_names = self.DPL.pattern_name_list
        assert set(["Net1_1", "Net2_1", "Net3_1"]).issubset(set(pattern_names))
    
    def test_add_pattern(self):
        pattern_name = "New_Pattern"
        entry = {"name": "New_Name",
                 "category": "New_Category",
                 "description": "New_Desription",
                 "citation": "New_Citation",
                 "start_clocktime": 0,
                 "pattern_timestep": 3600,
                 "wrap": True,
                 "multipliers": [int(i) for i in range(25)], #np.linspace(0, 24, 1), #list(np.arange(0,25,1))
                 }

        self.DPL.add_pattern(pattern_name, entry)
        pat = self.DPL.get_pattern("New_Pattern")
        series = self.DPL.to_Series("New_Pattern")
        
        expected = pd.Series(data=np.arange(0,25,1), index=np.arange(0,86401,3600))
        assert_series_equal(series, expected, check_dtype=False)
        
        # Try adding the pattern again (AssertionError, name already exists)
        self.assertRaises(AssertionError, self.DPL.add_pattern, pattern_name, entry)

    def test_add_pulse_pattern(self):
        # Pulse: on between 3:00 and 6:00 (3), 14:00 and 20:00 (6) 
        self.DPL.add_pulse_pattern('Pulse', [3*3600,6*3600,14*3600,20*3600])
        self.DPL.add_pulse_pattern('Pulse_invert', [3*3600,6*3600,14*3600,20*3600], 
                                   invert=True)
        
        pulse = self.DPL.to_Series('Pulse')
        assert pulse.mean() == 9/24
        
        pulse_inverse = self.DPL.to_Series('Pulse_invert')
        assert pulse_inverse.mean() == 15/24
        
        self.DPL.add_combined_pattern('Combined_pulse', 
                                      ['Pulse', 'Pulse_invert'], 
                                      combine='Overlap', 
                                      weights=None, 
                                      durations=[7*24*3600], 
                                      pattern_timestep=3600, 
                                      start_clocktime=0,
                                      wrap=True, normalize=False)
        combined_pulse = self.DPL.to_Series('Combined_pulse')
        #self.DPL.plot_patterns(names=['Pulse', 'Pulse_invert', 'Combined_pulse'])
        
        assert combined_pulse.mean() == 1
        assert combined_pulse.std() == 0
    
    def test_binary_pattern(self):
        """
        Pattern.binary_pattern and DPL.add_pulse_pattern share functionality.
        TODO Pattern.binary_pattern should be deprecated
        """
        wn = wntr.network.WaterNetworkModel(ex_datadir+'/Net3.inp')
        duration = 24*3600
        pattern_timestep = wn.options.time.pattern_timestep
        
        binary_pattern = wntr.network.elements.Pattern.binary_pattern(
            "binary_pattern", 3*3600, 6*3600, pattern_timestep, duration)
        wn.add_pattern(binary_pattern.name, binary_pattern)
        
        self.DPL.add_pulse_pattern('pulse_pattern', [3*3600,6*3600], 
                                   duration, pattern_timestep)
        pulse_pattern = self.DPL.to_Pattern('pulse_pattern')
        wn.add_pattern(pulse_pattern.name, pulse_pattern)
        
        assert all(binary_pattern.multipliers == pulse_pattern.multipliers)
        
    def test_add_gaussian_pattern(self):
        self.DPL.add_gaussian_pattern('Gauss', 3.5*24*3600, 12*3600, 
                                      duration=7*24*3600)
        gaussian = self.DPL.to_Series('Gauss')
        
        assert gaussian.idxmax() == 3.5*24*3600
    
    def test_add_triangular_pattern(self):
        self.DPL.add_triangular_pattern('Triangular', 2*3600, 12*3600, 18*3600, 
                                        normalize=True)
        triangle = self.DPL.to_Series('Triangular')
        
        assert triangle.idxmax() == 12*3600
        assert np.all(np.diff(triangle.loc[0:12*3600]) >= 0) # is monotonically increasing
        assert np.all(np.diff(triangle.loc[12*3600::]) <= 0) # is monotonically decreasing
    
    def test_add_combined_pattern_overlap(self):
        self.DPL.add_combined_pattern('Combined_overlap', 
                                      ['Net1_1', 'Net3_1'], 
                                      combine='Overlap', 
                                      weights=None, 
                                      durations=[7*24*3600], 
                                      pattern_timestep=3600, 
                                      start_clocktime=0,
                                      wrap=True, normalize=False)
        #self.DPL.plot_patterns(names=['Net1_1', 'Net3_1', 'Combined_overlap'])
        
        net1 = self.DPL.to_Pattern('Net1_1')
        net3 = self.DPL.to_Pattern('Net3_1')
        combined = self.DPL.to_Pattern('Combined_overlap')
        
        assert len(combined.multipliers) == 168
        self.assertAlmostEqual(combined.at(24*3600), net1.at(24*3600) + net3.at(24*3600), 3)
        
    def test_add_combined_pattern_sequential(self):
        self.DPL.add_combined_pattern('Combined_sequential', 
                                      ['Net1_1', 'Net3_1'], 
                                      combine='Sequential', 
                                      weights=None, 
                                      durations=[3.5*3600*24, 3.5*3600*24], 
                                      pattern_timestep=3600, 
                                      start_clocktime=0,
                                      wrap=True, normalize=False)
        #self.DPL.plot_patterns(names=['Net1_1', 'Net3_1', 'Combined_sequential'])
        
        net1 = self.DPL.to_Pattern('Net1_1')
        net3 = self.DPL.to_Pattern('Net3_1')
        combined = self.DPL.to_Pattern('Combined_sequential')
        
        assert len(combined.multipliers) == 168
        self.assertAlmostEqual(combined.at(2*24*3600), net1.at(2*24*3600), 3)
        self.assertAlmostEqual(combined.at(5*24*3600), net3.at(5*24*3600), 3)
    
    def test_remove_pattern(self):
        pattern_name = "Constant"

        self.DPL.remove_pattern(pattern_name)
        
        pattern_names = self.DPL.pattern_name_list
        assert pattern_name not in set(pattern_names)
        
        # Try removing the pattern again (AssertionError, name does not exists)
        self.assertRaises(AssertionError, self.DPL.remove_pattern, pattern_name)
    
    def test_copy_pattern(self):
        # Copy pattern
        self.DPL.copy_pattern('Net1_1', 'Net1_2')
        
        pat1 = self.DPL.to_Series('Net1_1')
        pat2 = self.DPL.to_Series('Net1_2')
        
        assert_series_equal(pat1, pat2)
        
    def test_read_write(self):
        num_patterns = len(self.DPL.pattern_name_list)
        self.DPL.write_json('New_demand_pattern_library.json')
        DPL2 = DemandPatternLibrary('New_demand_pattern_library.json')
        assert len(DPL2.pattern_name_list) == num_patterns
        
    def test_filter_by_category(self):
        # Filter patterns by category
        reidential_patterns = self.DPL.filter_by_category('Residential')
        commercial_patterns = self.DPL.filter_by_category('Commercial')
        indistrial_patterns = self.DPL.filter_by_category('Industrial')
        none_patterns = self.DPL.filter_by_category(None)
        
        assert 'Micropolis_2' in list(pd.DataFrame(reidential_patterns)['name'])
        assert 'Micropolis_1' in list(pd.DataFrame(commercial_patterns)['name'])
        assert 'Micropolis_3' in list(pd.DataFrame(indistrial_patterns)['name'])
        assert 'Null' in list(pd.DataFrame(none_patterns)['name'])
        
    def test_normalize(self):
        pattern_name = 'Net3_1'

        multipliers = self.DPL.get_pattern(pattern_name)['multipliers']
        mean_val = np.mean(multipliers)
        std_val = np.std(multipliers)
        
        # check that the original mean was not equal to 1
        self.assertNotAlmostEqual(mean_val, 1, 3) 
        
        self.DPL.normalize_pattern(pattern_name)
        multipliers = self.DPL.get_pattern(pattern_name)['multipliers']
        mean_val_norm = np.mean(multipliers)
        std_val_norm = np.std(multipliers)

        self.assertAlmostEqual(mean_val_norm, 1, 3)
        
    def test_resample_add_noise(self):
        val = 0.25
        
        self.DPL.add_gaussian_pattern('Gaussian', 12*3600, 5*3600, normalize=True)
        
        pat = self.DPL.get_pattern('Gaussian')
        length = len(pat['multipliers'])
        duration = len(pat['multipliers'])*pat['pattern_timestep']
        assert length == 24
        assert duration == 24*3600
        
        # Create a longer timeseries
        self.DPL.resample_multipliers('Gaussian', duration=7*24*3600, pattern_timestep=60, start_clocktime=0)
        
        pat = self.DPL.get_pattern('Gaussian')
        duration = len(pat['multipliers'])*pat['pattern_timestep']
        assert duration == 7*24*3600
        
        # Copy pattern and apply noise
        self.DPL.copy_pattern('Gaussian', 'Gaussian_with_noise')
        self.DPL.apply_noise('Gaussian_with_noise', val, seed=123)
        
        mult1 = self.DPL.get_pattern('Gaussian')['multipliers']
        mult2 = self.DPL.get_pattern('Gaussian_with_noise')['multipliers']
        diff = np.array(mult1) - np.array(mult2)
        mean_diff = np.mean(diff)
        std_diff = np.std(diff)
        
        self.assertAlmostEqual(mean_diff, 0, 2)
        self.assertAlmostEqual(std_diff, val, 2)
        
    def test_to_Pattern(self):
        pattern_name = 'Net3_1'
        mult = self.DPL.get_pattern(pattern_name)['multipliers']
        
        # Convert to a WNTR Pattern object
        pattern = self.DPL.to_Pattern(pattern_name)
        
        MEA = np.mean(np.abs(pattern.multipliers -  np.array(mult)))
        
        assert pattern.name == pattern_name
        self.assertAlmostEqual(MEA, 0, 6)
        
    def test_to_Series(self):
        pattern_name = 'Net3_1'
        mult = self.DPL.get_pattern(pattern_name)['multipliers']
        
        # Convert to a Pandas Series and change time parameters, this could be used to 
        # update the pattern or create a new pattern
        series1 = self.DPL.to_Series('Net3_1')
        series2 = self.DPL.to_Series('Net3_1', duration=4*24*3600)
        
        MEA = np.mean(np.abs(series1.values -  np.array(mult)))
        
        self.assertAlmostEqual(MEA, 0, 6)
        assert series2.shape[0] == 4*24
        
    def test_add_to_wn(self):
        
        # Create a water network model
        wn = wntr.network.WaterNetworkModel(ex_datadir+'/Net1.inp')
        wn.options.hydraulic.demand_model='PDD'
        wn.options.hydraulic.required_pressure = 20
        sim = wntr.sim.EpanetSimulator(wn)
        results = sim.run_sim()
        demand0 = results.node['demand']
        expected_demand0 = wntr.metrics.expected_demand(wn)
        
        # Get demands associated with a junction
        junction = wn.get_node('11')
        assert wn.num_patterns == 1
        assert len(junction.demand_timeseries_list) == 1

        # Modify the base value and pattern of the original demand
        #junction.demand_timeseries_list[0].base_value = 6e-5
        #junction.demand_timeseries_list[0].pattern_name = '1'
        junction.demand_timeseries_list[0].category = 'A'
        
        # Resample Net2_1 pattern to pattern timestep from Net1
        wn_pattern_timestep = wn.options.time.pattern_timestep
        pattern_timestep = self.DPL.get_pattern('Net2_1')['pattern_timestep']
        assert wn_pattern_timestep != pattern_timestep
        self.DPL.copy_pattern('Net2_1', 'Net2_1_resampled')
        multipliers = self.DPL.resample_multipliers('Net2_1_resampled', 
                                                    duration=3*24*3600,
                                                    pattern_timestep=wn_pattern_timestep, 
                                                    start_clocktime=0)
        self.DPL.normalize_pattern('Net2_1_resampled')
        #self.DPL.plot_patterns(names=['Net2_1_resampled', 'Net2_1'])
    
        # Add a new pattern from the pattern library, then add a demand
        pattern = self.DPL.to_Pattern('Net2_1_resampled', wn.options.time)
        wn.add_pattern('Net2_1_resample', pattern)
        junction.add_demand(base=5e-3, pattern_name='Net2_1_resample', category='B')
        assert pattern.name == 'Net2_1_resampled'
        assert wn.num_patterns == 2
        assert len(junction.demand_timeseries_list) == 2

        # Add another pattern from a list, then add a demand
        wn.add_pattern('New', [2,2,2,0,0,0,2,0,1,1,1,2])
        junction.add_demand(base=2e-3, pattern_name='New', category='C')
        assert wn.num_patterns == 3
        assert len(junction.demand_timeseries_list) == 3
        
        expected_demand1 = wntr.metrics.expected_demand(wn)
        
        sim = wntr.sim.EpanetSimulator(wn)
        results = sim.run_sim()
        demand1 = results.node['demand']
        expected_demand1 = wntr.metrics.expected_demand(wn)
        
        #fig, ax = plt.subplots()
        #expected_demand0['11'].plot(ax=ax)
        #expected_demand1['11'].plot(ax=ax)
        #demand0['11'].plot(ax=ax)
        #demand1['11'].plot(ax=ax)
        
        # Test that demand1 is greater than demand0
        assert all(expected_demand1['11'] > expected_demand0['11'])
        assert all(demand1['11'] > demand0['11'])
        
    def test_plot_pattern(self):
        filename = abspath(join(testdir, "plot_pattern.png"))
        if isfile(filename):
            os.remove(filename)
        
        # Plot patterns
        ax1 = self.DPL.plot_patterns(names=['Net1_1', 'Net2_1', 'Net3_1'])
        
        plt.savefig(filename, format="png")
        plt.close()


if __name__ == "__main__":
    unittest.main()



"""WNTR's algebraic modeling language module (SWIG)."""

from .expr import Var, Param, exp, log, sin, cos, tan, asin, acos, atan, inequality, sign, abs, value, ConditionalExpression
from .aml import Model, ParamDict, VarDict, ConstraintDict, Constraint




import unittest
import warnings
from os.path import abspath, dirname, join

import numpy as np
import pandas as pd
import wntr
from wntr.network.controls import Control, Rule

try:
    import geopandas as gpd
    has_geopandas = True
except ModuleNotFoundError:
    gpd = None
    has_geopandas = False
    
testdir = dirname(abspath(str(__file__)))
test_network_dir = join(testdir, "networks_for_testing")
test_data_dir = join(testdir, "data_for_testing")
ex_datadir = join(testdir, "..", "..", "examples", "networks")


class TestNetworkCreation(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        import wntr

        self.wntr = wntr

        inp_file = join(ex_datadir, "Net6.inp")
        self.wn = self.wntr.network.WaterNetworkModel(inp_file)

    @classmethod
    def tearDownClass(self):
        pass

    def test_num_junctions(self):
        self.assertEqual(self.wn.num_junctions, 3323)

    def test_num_reservoirs(self):
        self.assertEqual(self.wn.num_reservoirs, 1)

    def test_num_tanks(self):
        self.assertEqual(self.wn.num_tanks, 32)

    def test_num_pipes(self):
        self.assertEqual(self.wn.num_pipes, 3829)

    def test_num_pumps(self):
        self.assertEqual(self.wn.num_pumps, 61)

    def test_num_valves(self):
        self.assertEqual(self.wn.num_valves, 2)

    def test_num_nodes(self):
        self.assertEqual(self.wn.num_nodes, 3323 + 1 + 32)

    def test_num_links(self):
        self.assertEqual(self.wn.num_links, 3829 + 61 + 2)

    def test_junction_attr(self):
        j = self.wn.get_node("JUNCTION-18")
        # self.assertAlmostEqual(j.base_demand, 48.56/60.0*0.003785411784)
        self.assertEqual(j.demand_timeseries_list[0].pattern_name, "PATTERN-2")
        self.assertAlmostEqual(j.elevation, 80.0 * 0.3048)

    def test_reservoir_attr(self):
        j = self.wn.get_node("RESERVOIR-3323")
        self.assertAlmostEqual(j.head_timeseries.base_value, 27.45 * 0.3048)
        self.assertEqual(j.head_timeseries.pattern_name, None)


class TestNetworkMethods(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        import wntr
        from wntr.network.controls import Control, ControlAction
        from wntr.network.model import Junction, Pipe, Pump, Reservoir, Tank, Valve

        self.Junction = Junction
        self.Tank = Tank
        self.Reservoir = Reservoir
        self.Pipe = Pipe
        self.Pump = Pump
        self.Valve = Valve
        self.wntr = wntr
        self.ControlAction = ControlAction
        self.TimeControl = Control._time_control

    @classmethod
    def tearDownClass(self):
        pass

    def test_add_junction(self):
        wn = self.wntr.network.WaterNetworkModel()
        wn.add_pattern("pattern1", [1])
        wn.add_junction("j1", 150, "pattern1", 15)
        j = wn.get_node("j1")
        self.assertEqual(j._name, "j1")
        # self.assertEqual(j.base_demand, 150.0)
        self.assertEqual(j.demand_timeseries_list[0].pattern_name, "pattern1")
        self.assertEqual(j.elevation, 15.0)
        self.assertEqual(list(wn.node_name_list), ["j1"])
        # self.assertEqual(type(j.base_demand), float)
        self.assertEqual(type(j.elevation), float)

    def test_add_tank(self):
        wn = self.wntr.network.WaterNetworkModel()
        wn.add_tank("t1", 15, 75, 0, 100, 10, 0)
        n = wn.get_node("t1")
        self.assertEqual(n._name, "t1")
        self.assertEqual(n.elevation, 15.0)
        self.assertEqual(n.init_level, 75.0)
        self.assertEqual(n.min_level, 0.0)
        self.assertEqual(n.max_level, 100.0)
        self.assertEqual(n.diameter, 10.0)
        self.assertEqual(n.min_vol, 0.0)
        self.assertEqual(list(wn.node_name_list), ["t1"])
        self.assertEqual(type(n.elevation), float)
        self.assertEqual(type(n.init_level), float)
        self.assertEqual(type(n.min_level), float)
        self.assertEqual(type(n.max_level), float)
        self.assertEqual(type(n.diameter), float)
        self.assertEqual(type(n.min_vol), float)

    def test_add_reservoir(self):
        wn = self.wntr.network.WaterNetworkModel()
        wn.add_pattern("pattern1", [1])
        wn.add_reservoir("r1", 30, "pattern1")
        n = wn.get_node("r1")
        self.assertEqual(n._name, "r1")
        self.assertEqual(n.head_timeseries.base_value, 30.0)
        self.assertEqual(n.head_timeseries.pattern_name, "pattern1")
        self.assertEqual(list(wn.node_name_list), ["r1"])

    def test_add_pipe(self):
        wn = self.wntr.network.WaterNetworkModel()
        wn.add_junction("j1")
        wn.add_junction("j2")
        wn.add_pipe("p1", "j1", "j2", 1000, 1, 100, 0, "Open")
        l = wn.get_link("p1")
        self.assertEqual(l.name, "p1")
        self.assertEqual(l.start_node_name, "j1")
        self.assertEqual(l.end_node_name, "j2")
        self.assertEqual(l.initial_status, self.wntr.network.LinkStatus.Opened)
        self.assertEqual(l.length, 1000.0)
        self.assertEqual(l.diameter, 1.0)
        self.assertEqual(l.roughness, 100.0)
        self.assertEqual(l.minor_loss, 0.0)
        self.assertEqual(type(l.length), float)
        self.assertEqual(type(l.diameter), float)
        self.assertEqual(type(l.roughness), float)
        self.assertEqual(type(l.minor_loss), float)

    def test_add_pattern(self):
        wn = self.wntr.network.WaterNetworkModel()
        wn.add_junction("j1")
        wn.options.time.duration = 10
        wn.options.time.pattern_timestep = 1
        #        wn.add_pattern('pat1', start_time=2, end_time=4)
        pat1 = self.wntr.network.elements.Pattern.binary_pattern(
            "pat1", start_time=2, end_time=4, duration=10, step_size=1
        )
        wn.add_pattern("pat1", pat1)
        wn.add_pattern("pat2", [1, 2, 3, 4])

        pat1 = wn.get_pattern("pat1")
        pat2 = wn.get_pattern("pat2")

        self.assertEqual(
            pat1.multipliers.tolist(),
            [0.0, 0.0, 1.0, 1.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
        )
        self.assertEqual(pat2.multipliers.tolist(), [1, 2, 3, 4])

    def test_add_source(self):
        wn = self.wntr.network.WaterNetworkModel()
        wn.add_junction("j1")
        wn.options.time.duration = 10
        wn.options.time.pattern_timestep = 1
        pat1 = self.wntr.network.elements.Pattern.binary_pattern(
            "pat1",
            start_time=2,
            end_time=4,
            duration=10,
            step_size=wn.options.time.pattern_timestep,
        )
        wn.add_pattern("pat1", pat1)
        wn.add_source("s1", "j1", "SETPOINT", 100, "pat1")
        s = wn.get_source("s1")
        self.assertEqual(s.name, "s1")
        self.assertEqual(s.node_name, "j1")
        self.assertEqual(s.source_type, "SETPOINT")
        self.assertEqual(s.strength_timeseries.base_value, 100)
        self.assertEqual(s.strength_timeseries.pattern_name, "pat1")

    def test_add_pipe_with_cv(self):
        wn = self.wntr.network.WaterNetworkModel()
        wn.add_junction("j1")
        wn.add_junction("j2")
        wn.add_pipe("p1", "j1", "j2", 1000, 1, 100, 0, "OPEN", True)
        self.assertTrue(wn.get_link('p1').check_valve)

    def test_remove_pipe(self):
        wn = self.wntr.network.WaterNetworkModel()
        wn.add_junction("j1")
        wn.add_junction("j2")
        wn.add_junction("j3")
        wn.add_pipe("p2", "j1", "j3")
        wn.add_pipe("p1", "j1", "j2", initial_status=self.wntr.network.LinkStatus.CV)
        wn.remove_link("p1")
        link_list = [link_name for link_name, link in wn.links()]
        self.assertEqual(link_list, ["p2"])
        self.assertEqual(wn.num_pipes, 1)

    def test_remove_node(self):
        inp_file = join(ex_datadir, "Net6.inp")
        wn = self.wntr.network.WaterNetworkModel(inp_file)

        links_to_remove = wn.get_links_for_node("TANK-3326")
        for l in links_to_remove:
            wn.remove_link(l, with_control=True)
        wn.remove_node("TANK-3326", with_control=True)

        self.assertNotIn("TANK-3326", {name for name, node in wn.nodes()})
        self.assertNotIn("TANK-3326", wn.node_name_list)

        inp_file = join(test_network_dir, "conditional_controls_1.inp")
        wn = self.wntr.network.WaterNetworkModel(inp_file)

        tank1 = wn.get_node("tank1")
        action = self.ControlAction(tank1, "elevation", 55)
        control = self.TimeControl(wn, 3652, "SIM_TIME", False, action)
        wn.add_control("tank_control", control)

        controls_1 = {c_name for c_name, c in wn.controls()}

        links_to_remove = wn.get_links_for_node("tank1")
        for l in links_to_remove:
            wn.remove_link(l, with_control=True)
        wn.remove_node("tank1", with_control=True)

        controls_2 = {c_name for c_name, c in wn.controls()}

        self.assertTrue("tank_control" in controls_1)
        self.assertFalse("tank_control" in controls_2)

        self.assertNotIn("tank1", {name for name, node in wn.nodes()})
        self.assertNotIn("tank1", wn.node_name_list)
        expected_nodes = {"junction1", "res1"}
        self.assertSetEqual({name for name, node in wn.nodes()}, expected_nodes)

    def test_remove_controls_for_removing_link(self):
        inp_file = join(ex_datadir, "Net1.inp")
        wn = self.wntr.network.WaterNetworkModel(inp_file)

        control_action = self.wntr.network.ControlAction(
            wn.get_link("21"), "status", self.wntr.network.LinkStatus.Opened
        )
        control = self.wntr.network.controls.Control._conditional_control(
            wn.get_node("2"), "head", np.greater_equal, 10.0, control_action
        )
        wn.add_control("control_1", control)

        controls_1 = {c_name for c_name, c in wn.controls()}

        wn.remove_link("21", with_control=True)

        controls_2 = {c_name for c_name, c in wn.controls()}
        self.assertTrue("control_1" in controls_1)
        self.assertFalse("control_1" in controls_2)

    def test_nodes(self):
        wn = self.wntr.network.WaterNetworkModel()
        wn.add_junction("j1")
        wn.add_junction("j2")
        wn.add_tank("t1")
        wn.add_reservoir("r1")
        wn.add_reservoir("r2")
        wn.add_junction("j3")
        node_list = [name for name, node in wn.nodes()]
        node_list.sort()
        junction_list = [name for name, junction in wn.nodes(self.Junction)]
        junction_list.sort()
        tank_list = [name for name, tank in wn.nodes(self.Tank)]
        tank_list.sort()
        reservoir_list = [name for name, reservoir in wn.nodes(self.Reservoir)]
        reservoir_list.sort()
        self.assertEqual(node_list, ["j1", "j2", "j3", "r1", "r2", "t1"])
        self.assertEqual(junction_list, ["j1", "j2", "j3"])
        self.assertEqual(tank_list, ["t1"])
        self.assertEqual(reservoir_list, ["r1", "r2"])
        for name, node in wn.nodes():
            self.assertEqual(name, node._name)

    def test_links(self):
        wn = self.wntr.network.WaterNetworkModel()
        wn.add_junction("j1")
        wn.add_junction("j2")
        wn.add_tank("t1")
        wn.add_tank("t2")
        wn.add_reservoir("r1")
        wn.add_reservoir("r2")
        wn.add_junction("j3")
        wn.add_pipe("p1", "j1", "j2")
        wn.add_pipe("p2", "j1", "t1")
        wn.add_pipe("p3", "r1", "j1")
        wn.add_pump("pump1", "r2", "t2")
        wn.add_valve("v1", "j3", "j2")
        link_list = [name for name, link in wn.links()]
        link_list.sort()
        self.assertEqual(link_list, ["p1", "p2", "p3", "pump1", "v1"])
        pipe_list = [name for name, pipe in wn.links(self.Pipe)]
        pipe_list.sort()
        self.assertEqual(pipe_list, ["p1", "p2", "p3"])
        pump_list = [name for name, pump in wn.links(self.Pump)]
        pump_list.sort()
        self.assertEqual(pump_list, ["pump1"])
        valve_list = [name for name, valve in wn.links(self.Valve)]
        valve_list.sort()
        self.assertEqual(valve_list, ["v1"])
        for name, link in wn.links():
            self.assertEqual(name, link._link_name)

    def test_1_pt_head_curve(self):
        q2 = 10.0
        h2 = 20.0
        wn = self.wntr.network.WaterNetworkModel()
        wn.add_curve("curve1", "HEAD", [(q2, h2)])
        curve = wn.get_curve("curve1")
        wn.add_junction("j1")
        wn.add_junction("j2")
        wn.add_pump("p1", "j1", "j2", "HEAD", curve.name)
        link = wn.get_link("p1")
        a, b, c = link.get_head_curve_coefficients()
        q1 = 0.0
        h1 = 4.0 / 3.0 * h2
        q3 = 2.0 * q2
        h3 = 0.0

        def curve_fun(x):
            f = [1.0, 1.0, 1.0]
            f[0] = h1 - x[0] + x[1] * q1 ** x[2]
            f[1] = h2 - x[0] + x[1] * q2 ** x[2]
            f[2] = h3 - x[0] + x[1] * q3 ** x[2]
            return f

        X = [a, b, c]
        Y = curve_fun(X)

        self.assertAlmostEqual(Y[0], 0.0)
        self.assertAlmostEqual(Y[1], 0.0)
        self.assertAlmostEqual(Y[2], 0.0)

    def test_2_pt_head_curve(self):
        q1 = 0.0
        h1 = 1.0
        q2 = 1.0
        h2 = 0.0
        wn = self.wntr.network.WaterNetworkModel()
        wn.add_curve("curve1", "HEAD", [(q1, h1), (q2, h2)])
        curve = wn.get_curve("curve1")
        wn.add_junction("j1")
        wn.add_junction("j2")
        wn.add_pump("p1", "j1", "j2", "HEAD", curve.name)
        link = wn.get_link("p1")
        a, b, c = link.get_head_curve_coefficients()

        Y = [a - b, 0.5 - (a - b * (0.5) ** c), 1.0 - a]

        self.assertAlmostEqual(Y[0], 0.0)
        self.assertAlmostEqual(Y[1], 0.0)
        self.assertAlmostEqual(Y[2], 0.0)

    
    def test_3_pt_head_curve(self):
        q1 = 0.0
        h1 = 35.0
        q2 = 10.0
        h2 = 20.0
        q3 = 18.0
        h3 = 2.0
        wn = self.wntr.network.WaterNetworkModel()
        wn.add_curve("curve1", "HEAD", [(q1, h1), (q2, h2), (q3, h3)])
        curve = wn.get_curve("curve1")
        wn.add_junction("j1")
        wn.add_junction("j2")
        wn.add_pump("p1", "j1", "j2", "HEAD", curve.name)
        link = wn.get_link("p1")
        a, b, c = link.get_head_curve_coefficients()

        def curve_fun(x):
            f = [1.0, 1.0, 1.0]
            f[0] = h1 - x[0] + x[1] * q1 ** x[2]
            f[1] = h2 - x[0] + x[1] * q2 ** x[2]
            f[2] = h3 - x[0] + x[1] * q3 ** x[2]
            return f

        X = [a, b, c]
        Y = curve_fun(X)

        self.assertAlmostEqual(Y[0], 0.0)
        self.assertAlmostEqual(Y[1], 0.0)
        self.assertAlmostEqual(Y[2], 0.0)

    def test_multi_pt_head_curve(self):
        # pump_curves = pump_curves_for_testing() # change this to read in a csv file

        df = pd.read_csv(join(test_data_dir, "pump_practice_curves.csv"), skiprows=5)
        pump_curves = []
        for i in range(11):
            pump_curves.append(df[df["curve number"] == i].iloc[:, 1:3])

        # these are the least squares optimal curve coefficients for
        # pump_curves!
        expected_coef = [
            (95.2793750017631, 92.93210451708887, 1.7962733026912),
            (65.73126364420834, 754.0506268456613, 3.4783989343179305),
            (21.278376275311476, 4077.4535673165924, 2.358360722834581),
            (80.75954376866605, 7027.4864048631935, 4.082009299805731),
            (136.17943689581463, 829.0733504294045, 4.489091828720322),
            (127.24246902634025, 180.0189467345297, 3.932735276445609),
            (26.851737190843544, 546.1599637089113, 2.200609181978668),
            (49.44241876166496, 933337.4378918532, 2.7686910719067157),
            (34.29268375712387, 110108.09777459255, 3.522011383917406),
            (34.11693364265361, 170.07427285446153, 1.6314432418473557),
            (14.886248327061447, 3667.672645345962, 2.1488882176053345),
        ]

        # start an empty WNTR model and add dummy junctions so that
        # head pumps can be added
        wn = self.wntr.network.WaterNetworkModel()
        wn.add_junction("j1")
        wn.add_junction("j2")

        # check all of the pump_curves
        for idx, curve in enumerate(pump_curves):
            pump_name = "p" + str(idx + 1)
            curve_name = "c" + str(idx + 1)

            wn.add_curve(curve_name, "HEAD", curve.values)
            wn.add_pump(
                pump_name, "j1", "j2", pump_type="HEAD", pump_parameter=curve_name
            )
            pump = wn.get_link(pump_name)
            coef = pump.get_head_curve_coefficients()

            for i in range(2):
                error = abs(coef[i] - expected_coef[idx][i]) / expected_coef[idx][i]
                self.assertLess(error, 1e-6)
            # self.wntr.graphics.curve.plot_pump_curve(pump)
            # savefig(pump_name + "_" + cname + ".png")

    
    def test_multi_pt_head_curve_expected_error(self):
        # now test two error conditions using bad datasets.
        wn = self.wntr.network.WaterNetworkModel()
        wn.add_junction("j1")
        wn.add_junction("j2")

        # Negative coefficients
        data_points1 = [(x, x ** 2 + 2 * x + 3) for x in range(10)]
        wn.add_curve("curve1", "HEAD", data_points1)
        wn.add_pump("pump1", "j1", "j2", pump_type="HEAD", pump_parameter="curve1")
        pump = wn.get_link("pump1")
        with self.assertRaises(RuntimeError):
            pump.get_head_curve_coefficients()

        # Bad fit
        data_points2 = [(x, (-1) ** x * np.exp(-0.001 * x) * 100) for x in range(10)]
        wn.add_curve("curve2", "HEAD", data_points2)
        wn.add_pump("pump2", "j1", "j2", pump_type="HEAD", pump_parameter="curve2")
        pump = wn.get_link("pump2")
        with self.assertRaises(RuntimeError):
            pump.get_head_curve_coefficients()

    def test_get_links_for_node_all(self):
        wn = self.wntr.network.WaterNetworkModel()
        wn.add_junction("j1")
        wn.add_junction("j2")
        wn.add_junction("j3")
        wn.add_junction("j4")
        wn.add_junction("j5")
        wn.add_pipe("p1", "j1", "j2")
        wn.add_pipe("p2", "j1", "j4")
        wn.add_pipe("p3", "j1", "j5")
        wn.add_pipe("p4", "j3", "j2")
        wn.add_pipe("p5", "j4", "j3")
        l1 = wn.get_links_for_node("j1")
        l2 = wn.get_links_for_node("j2", "all")
        l3 = wn.get_links_for_node("j3", "all")
        l4 = wn.get_links_for_node("j4")
        l5 = wn.get_links_for_node("j5")
        l1.sort()
        l2.sort()
        l3.sort()
        l4.sort()
        l5.sort()
        self.assertEqual(l1, ["p1", "p2", "p3"])
        self.assertEqual(l2, ["p1", "p4"])
        self.assertEqual(l3, ["p4", "p5"])
        self.assertEqual(l4, ["p2", "p5"])
        self.assertEqual(l5, ["p3"])

    def test_get_links_for_node_in(self):
        wn = self.wntr.network.WaterNetworkModel()
        wn.add_junction("j1")
        wn.add_junction("j2")
        wn.add_junction("j3")
        wn.add_junction("j4")
        wn.add_junction("j5")
        wn.add_pipe("p1", "j1", "j2")
        wn.add_pipe("p2", "j1", "j4")
        wn.add_pipe("p3", "j1", "j5")
        wn.add_pipe("p4", "j3", "j2")
        wn.add_pipe("p5", "j4", "j3")
        l1 = wn.get_links_for_node("j1", "inlet")
        l2 = wn.get_links_for_node("j2", "inlet")
        l3 = wn.get_links_for_node("j3", "inlet")
        l4 = wn.get_links_for_node("j4", "inlet")
        l5 = wn.get_links_for_node("j5", "inlet")
        l1.sort()
        l2.sort()
        l3.sort()
        l4.sort()
        l5.sort()
        self.assertEqual(l1, [])
        self.assertEqual(l2, ["p1", "p4"])
        self.assertEqual(l3, ["p5"])
        self.assertEqual(l4, ["p2"])
        self.assertEqual(l5, ["p3"])

    def test_get_links_for_node_out(self):
        wn = self.wntr.network.WaterNetworkModel()
        wn.add_junction("j1")
        wn.add_junction("j2")
        wn.add_junction("j3")
        wn.add_junction("j4")
        wn.add_junction("j5")
        wn.add_pipe("p1", "j1", "j2")
        wn.add_pipe("p2", "j1", "j4")
        wn.add_pipe("p3", "j1", "j5")
        wn.add_pipe("p4", "j3", "j2")
        wn.add_pipe("p5", "j4", "j3")
        l1 = wn.get_links_for_node("j1", "outlet")
        l2 = wn.get_links_for_node("j2", "outlet")
        l3 = wn.get_links_for_node("j3", "outlet")
        l4 = wn.get_links_for_node("j4", "outlet")
        l5 = wn.get_links_for_node("j5", "outlet")
        l1.sort()
        l2.sort()
        l3.sort()
        l4.sort()
        l5.sort()
        self.assertEqual(l1, ["p1", "p2", "p3"])
        self.assertEqual(l2, [])
        self.assertEqual(l3, ["p4"])
        self.assertEqual(l4, ["p5"])
        self.assertEqual(l5, [])


epanet_unit_id = {
    "CFS": 0,
    "GPM": 1,
    "MGD": 2,
    "IMGD": 3,
    "AFD": 4,
    "LPS": 5,
    "LPM": 6,
    "MLD": 7,
    "CMH": 8,
    "CMD": 9,
}


class TestCase(unittest.TestCase):
    def test_Net1(self):
        inp_file = join(ex_datadir, "Net1.inp")
        wn = wntr.network.WaterNetworkModel(inp_file)
        G = wn.to_graph()

        node = G.nodes
        elevation = wn.query_node_attribute("elevation")
        # base_demand = wn.query_node_attribute('base_demand')
        edge = G.adj
        diameter = wn.query_link_attribute("diameter")
        length = wn.query_link_attribute("length")

        # Data from the INP file, converted using flowunits
        expected_node = {
            "11": {"type": "Junction", "pos": (30.0, 70.0)},
            "10": {"type": "Junction", "pos": (20.0, 70.0)},
            "13": {"type": "Junction", "pos": (70.0, 70.0)},
            "12": {"type": "Junction", "pos": (50.0, 70.0)},
            "21": {"type": "Junction", "pos": (30.0, 40.0)},
            "22": {"type": "Junction", "pos": (50.0, 40.0)},
            "23": {"type": "Junction", "pos": (70.0, 40.0)},
            "32": {"type": "Junction", "pos": (50.0, 10.0)},
            "31": {"type": "Junction", "pos": (30.0, 10.0)},
            "2": {"type": "Tank", "pos": (50.0, 90.0)},
            "9": {"type": "Reservoir", "pos": (10.0, 70.0)},
        }

        expected_elevation = {
            "11": 710.0,
            "10": 710.0,
            "13": 695.0,
            "12": 700.0,
            "21": 700.0,
            "22": 695.0,
            "23": 690.0,
            "32": 710.0,
            "31": 700.0,
            "2": 850.0,
        }
        expected_elevation = wntr.epanet.util.HydParam.Elevation._to_si(
            wn._inpfile.flow_units, expected_elevation
        )

        expected_base_demand = {
            "11": 150,
            "10": 0,
            "13": 100,
            "12": 150,
            "21": 150,
            "22": 200,
            "23": 150,
            "32": 100,
            "31": 100,
        }
        expected_base_demand = wntr.epanet.util.HydParam.Demand._to_si(
            wn._inpfile.flow_units, expected_base_demand
        )

        expected_edge = {
            "11": {"12": {"11": {"type": "Pipe"}}, "21": {"111": {"type": "Pipe"}}},
            "10": {"11": {"10": {"type": "Pipe"}}},
            "13": {"23": {"113": {"type": "Pipe"}}},
            "12": {"13": {"12": {"type": "Pipe"}}, "22": {"112": {"type": "Pipe"}}},
            "21": {"31": {"121": {"type": "Pipe"}}, "22": {"21": {"type": "Pipe"}}},
            "22": {"32": {"122": {"type": "Pipe"}}, "23": {"22": {"type": "Pipe"}}},
            "23": {},
            "32": {},
            "31": {"32": {"31": {"type": "Pipe"}}},
            "2": {"12": {"110": {"type": "Pipe"}}},
            "9": {"10": {"9": {"type": "Pump"}}},
        }

        expected_diameter = {
            "11": 14.0,
            "111": 10.0,
            "10": 18.0,
            "113": 8.0,
            "12": 10.0,
            "112": 12.0,
            "121": 8.0,
            "21": 10.0,
            "122": 6.0,
            "22": 12.0,
            "31": 6.0,
            "110": 18.0,
        }
        expected_diameter = wntr.epanet.util.HydParam.PipeDiameter._to_si(
            wn._inpfile.flow_units, expected_diameter
        )

        expected_length = {
            "11": 5280.0,
            "111": 5280.0,
            "10": 10530.0,
            "113": 5280.0,
            "12": 5280.0,
            "112": 5280.0,
            "121": 5280.0,
            "21": 5280.0,
            "122": 5280.0,
            "22": 5280.0,
            "31": 5280.0,
            "110": 200.0,
        }
        expected_length = wntr.epanet.util.HydParam.Length._to_si(
            wn._inpfile.flow_units, expected_length
        )

        self.assertDictEqual(dict(node), expected_node)
        self.assertDictEqual(dict(elevation), expected_elevation)
        # self.assertDictEqual(base_demand, expected_base_demand)

        self.assertDictEqual(dict(edge), expected_edge)
        self.assertDictEqual(dict(diameter), expected_diameter)
        self.assertDictEqual(dict(length), expected_length)

    def test_query_node_attribute(self):
        inp_file = join(ex_datadir, "Net1.inp")
        wn = wntr.network.WaterNetworkModel(inp_file)

        elevation = 213.36  # 700*float(units.ft/units.m) # ft to m
        nodes = wn.query_node_attribute("elevation", np.less, elevation)

        expected_nodes = set(["13", "22", "23"])

        self.assertSetEqual(set(nodes.keys()), expected_nodes)

    def test_query_pipe_attribute(self):
        inp_file = join(ex_datadir, "Net1.inp")
        wn = wntr.network.WaterNetworkModel(inp_file)

        length = 1609.344  # 5280*float(units.ft/units.m) # ft to m
        pipes = wn.query_link_attribute("length", np.greater, length)

        expected_pipes = set(["10"])

        self.assertSetEqual(set(pipes.keys()), expected_pipes)

    def test_nzd_nodes(self):
        inp_file = join(ex_datadir, "Net1.inp")
        wn = wntr.network.WaterNetworkModel(inp_file)

        nzd_nodes = []
        for name, node in wn.junctions():
            demand = 0
            for ts in np.arange(
                0, wn.options.time.duration, wn.options.time.report_timestep
            ):
                demand += node.demand_timeseries_list.at(ts)
            if demand > 0:
                nzd_nodes.append(name)

        expected_nodes = set(["11", "13", "12", "21", "22", "23", "32", "31"])

        self.assertSetEqual(set(nzd_nodes), expected_nodes)

    def test_name_list(self):
        inp_file = join(ex_datadir, "Net3.inp")
        wn = wntr.network.WaterNetworkModel(inp_file)

        self.assertIn("10", wn.junction_name_list)
        self.assertIn("1", wn.tank_name_list)
        self.assertIn("River", wn.reservoir_name_list)
        self.assertIn("20", wn.pipe_name_list)
        self.assertIn("10", wn.pump_name_list)
        self.assertEqual(0, len(wn.valve_name_list))
        self.assertIn("1", wn.pattern_name_list)
        self.assertIn("1", wn.curve_name_list)
        self.assertEqual(0, len(wn.source_name_list))
        #    self.assertEqual(0, len(wn._demand_name_list))
        self.assertIn("control 1", wn.control_name_list)

    def test_add_get_remove_num(self):
        inp_file = join(ex_datadir, "Net3.inp")
        wn = wntr.network.WaterNetworkModel(inp_file)

        wn.add_junction("new_junc")
        wn.get_node("new_junc")

        wn.add_tank("new_tank")
        wn.get_node("new_tank")

        wn.add_reservoir("new_reservoir")
        wn.get_node("new_reservoir")

        wn.add_pipe("new_pipe", "139", "131")
        wn.get_link("new_pipe")

        wn.add_pump("new_pump", "139", "131")
        wn.get_link("new_pump")

        wn.add_valve("new_valve", "139", "131")
        wn.get_link("new_valve")

        wn.add_pattern("new_pattern", [])
        wn.get_pattern("new_pattern")

        wn.add_curve("new_curve", "HEAD", [])
        wn.get_curve("new_curve")

        wn.add_source("new_source", "new_junc", "CONCEN", 1, "new_pattern")
        wn.get_source("new_source")

        nums = [
            wn.num_junctions,
            wn.num_tanks,
            wn.num_reservoirs,
            wn.num_pipes,
            wn.num_pumps,
            wn.num_valves,
            wn.num_patterns,
            wn.num_curves,
            wn.num_sources,
        ]
        expected = [93, 4, 3, 118, 3, 1, 6, 3, 1]
        self.assertListEqual(nums, expected)

        # Verify that runtime errors occur when there is a node/pattern still in use
        self.assertRaises(RuntimeError, wn.remove_node, "new_junc")
        self.assertRaises(RuntimeError, wn.remove_pattern, "1")

        wn.remove_source("new_source")
        wn.remove_curve("new_curve")
        wn.remove_pattern("new_pattern")
        wn.remove_link("new_pipe")
        wn.remove_link("new_pump")
        wn.remove_link("new_valve")
        wn.remove_node("new_junc")
        wn.remove_node("new_tank")
        wn.remove_node("new_reservoir")

        nums = [
            wn.num_junctions,
            wn.num_tanks,
            wn.num_reservoirs,
            wn.num_pipes,
            wn.num_pumps,
            wn.num_valves,
            wn.num_patterns,
            wn.num_curves,
            wn.num_sources,
        ]
        expected = [92, 3, 2, 117, 2, 0, 5, 2, 0]
        self.assertListEqual(nums, expected)

    def test_describe(self):
        inp_file = join(ex_datadir, "Net3.inp")
        wn = wntr.network.WaterNetworkModel(inp_file)
        d0 = wn.describe(0)
        self.assertDictEqual(
            d0,
            {
                "Nodes": 97,
                "Links": 119,
                "Patterns": 5,
                "Curves": 2,
                "Sources": 0,
                "Controls": 18,
            },
        )

        d1 = wn.describe(1)
        self.assertDictEqual(
            d1,
            {
                "Nodes": {"Junctions": 92, "Tanks": 3, "Reservoirs": 2},
                "Links": {"Pipes": 117, "Pumps": 2, "Valves": 0},
                "Patterns": 5,
                "Curves": {"Pump": 2, "Efficiency": 0, "Headloss": 0, "Volume": 0},
                "Sources": 0,
                "Controls": 18,
            },
        )

    def test_assign_demand(self):

        inp_file = join(ex_datadir, "Net3.inp")
        wn = wntr.network.WaterNetworkModel(inp_file)

        demands0 = wntr.metrics.expected_demand(wn)
        pattern_name0 = wn.get_node("10").demand_timeseries_list[0].pattern_name

        wn.options.hydraulic.demand_multiplier = 1.5
        demands1 = wntr.metrics.expected_demand(wn)

        self.assertEqual(pattern_name0, "1")
        self.assertEqual(len(wn.pattern_name_list), 5)  # number of original patterns
        self.assertLess(abs((demands1 / demands0).max().max() - 1.5), 0.000001)

        sim1 = wntr.sim.EpanetSimulator(wn)
        results1 = sim1.run_sim()
        demands_sim1 = results1.node["demand"].loc[:, wn.junction_name_list]

        ### re-assign demands to be 2 times the original demands
        wn.assign_demand(demands1 * 2, pattern_prefix="ResetDemand1_")

        demands2 = wntr.metrics.expected_demand(wn)
        pattern_name = wn.get_node("10").demand_timeseries_list[0].pattern_name

        sim = wntr.sim.EpanetSimulator(wn)
        results2 = sim.run_sim()
        demands_sim2 = results2.node["demand"].loc[:, wn.junction_name_list]

        self.assertEqual(pattern_name, "ResetDemand1_10")
        self.assertEqual(len(wn.pattern_name_list), wn.num_junctions + 5)
        self.assertLess(abs((demands2 / demands1).max().max() - 2), 0.000001)
        self.assertLess(abs((demands_sim2 / demands_sim1).max().max() - 2), 0.01)

        ### re-assign demands using results from the simulation
        wn.assign_demand(demands_sim2, pattern_prefix="ResetDemand2_")

        demands2 = wntr.metrics.expected_demand(wn)
        pattern_name = wn.get_node("10").demand_timeseries_list[0].pattern_name

        sim = wntr.sim.EpanetSimulator(wn)
        results2 = sim.run_sim()
        demands_sim2 = results2.node["demand"].loc[:, wn.junction_name_list]

        self.assertEqual(pattern_name, "ResetDemand2_10")
        self.assertEqual(len(wn.pattern_name_list), 2 * wn.num_junctions + 5)
        self.assertLess(abs((demands2 / demands1).max().max() - 2), 0.01)
        self.assertLess(abs((demands_sim2 / demands_sim1).max().max() - 2), 0.01)

    def test_convert_controls_to_rules(self):

        inp_file = join(ex_datadir, "Net3.inp")
        wn = wntr.network.WaterNetworkModel(inp_file)
        wn.options.time.rule_timestep = (
            1  # For an exact match, the rule timestep must be very small
        )

        nControls = 0
        nRules = 0
        for name, control in wn.controls():
            if isinstance(control, Control):
                nControls = nControls + 1
            elif isinstance(control, Rule):
                nRules = nRules + 1
        assert nControls == 18
        assert nRules == 0

        sim = wntr.sim.EpanetSimulator(wn)
        results1 = sim.run_sim()

        wn.convert_controls_to_rules(priority=3)

        nControls = 0
        nRules = 0
        for name, control in wn.controls():
            if isinstance(control, Control):
                nControls = nControls + 1
            elif isinstance(control, Rule):
                nRules = nRules + 1
        assert nControls == 0
        assert nRules == 18

        sim = wntr.sim.EpanetSimulator(wn)
        results2 = sim.run_sim()

        # (results1.node['pressure'] - results2.node['pressure']).plot()

        for node_name, node in wn.nodes():
            for t in results1.node["pressure"].index:
                self.assertLess(
                    abs(
                        results1.node["pressure"].at[t, node_name]
                        - results2.node["pressure"].at[t, node_name]
                    ),
                    0.001,
                )


class TestNetworkIO_Dict(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        import wntr

        self.wntr = wntr

        self.inp_file = join(ex_datadir, "Net3.inp")
        self.inp_files = [join(ex_datadir, f) for f in ["Net1.inp", "Net2.inp", "Net3.inp", "Net6.inp", "ky10.inp"]]

    @classmethod
    def tearDownClass(self):
        pass

    def test_dict_roundtrip(self):
        for inp_file in self.inp_files:
            wn = self.wntr.network.WaterNetworkModel(inp_file)
            A = wn.to_dict()
            B = self.wntr.network.from_dict(A)
            assert(wn._compare(B))

    def test_json_roundtrip(self):
        for inp_file in self.inp_files:
            wn = self.wntr.network.WaterNetworkModel(inp_file)
            wn.convert_controls_to_rules()
            self.wntr.network.write_json(wn, 'temp.json')
            B = self.wntr.network.read_json('temp.json')
            assert(wn._compare(B))

    def test_json_pattern_dump(self):
        wn = wntr.network.WaterNetworkModel()
        wn.add_pattern('pat0', [0,1,0,1,0,1,0])
        self.wntr.network.write_json(wn, f'temp.json')
        
    def test_dict_with_leak(self):
        # This covers a bug where writing controls to a dictionary broke if a leak was added
        wn = wntr.network.WaterNetworkModel(self.inp_file)
        junction_name = wn.junction_name_list[0]
        junction = wn.get_node(junction_name)
        junction.add_leak(wn, area=1, start_time=0, end_time=3600)
        wn_dict = wn.to_dict()
        wn2 = wntr.network.from_dict(wn_dict)
        
        # check leak controls and node properties
        start_control = wn.get_control(wn.control_name_list[18])
        start_control2 = wn2.get_control(wn2.control_name_list[18])
        assert str(start_control) == str(start_control2)
        
        end_control = wn.get_control(wn.control_name_list[19])
        end_control2 = wn2.get_control(wn2.control_name_list[19])
        assert str(end_control) == str(end_control2)

        junction2 = wn2.get_node(junction_name)
        assert junction2._leak
        assert junction2._leak_area == 1
        assert junction2._leak_discharge_coeff == 0.75

    def test_pump_efficiencey_curve(self):
        wn = wntr.network.WaterNetworkModel()
        wn.add_junction('J1')
        wn.add_junction('J2')
        wn.add_pump('P1','J1','J2')
        pump = wn.get_link('P1')
        wn.add_curve('efficiency_curve', 'EFFICIENCY',[(0,0),(1,1)])
        pump.efficiency_curve_name = 'efficiency_curve'


        wn_dict = wn.to_dict()
        wn2 = wntr.network.from_dict(wn_dict)

        efficiency_curve = wn2.get_curve('efficiency_curve')

        assert efficiency_curve is not None
        assert efficiency_curve.points == [(0, 0), (1, 1)]

        assert pump.efficiency_curve_name == 'efficiency_curve'
        assert isinstance(pump.efficiency_curve, wntr.network.Curve)
        assert pump.efficiency_curve.name == 'efficiency_curve'


@unittest.skipIf(not has_geopandas,
                 "Cannot test GIS capabilities: geopandas is missing")
class TestNetworkIO_GIS(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        import wntr

        self.wntr = wntr

        inp_file = join(ex_datadir, "Net6.inp")
        self.inp_files = [join(ex_datadir, f) for f in ["Net1.inp", "Net2.inp", "Net3.inp", "Net6.inp"]]

    @classmethod
    def tearDownClass(self):
        pass
    
    def test_gis_additional_attributes(self):
        wn = self.wntr.network.WaterNetworkModel(self.inp_files[0])
        A = wn.to_gis(pumps_as_points=True, valves_as_points=True)
        A.junctions['new'] = range(wn.num_junctions)
        A.junctions['pressure_exponent'] = 0.45
        A.tanks['new'] = 4
        A.reservoirs['new'] = 5
        A.pipes['new'] = 6
        A.pumps['new'] = 7
        A.valves['new'] = 8
        B = self.wntr.network.from_gis(A)

        node_attr = B.query_node_attribute('new')
        assert((node_attr[wn.junction_name_list] == range(wn.num_junctions)).all())
        assert((node_attr[wn.tank_name_list] == 4).all())
        assert((node_attr[wn.reservoir_name_list] == 5).all())
        
        link_attr = B.query_link_attribute('new')
        assert((link_attr[wn.pipe_name_list] == 6).all())
        assert((link_attr[wn.pump_name_list] == 7).all())
        assert((link_attr[wn.valve_name_list] == 8).all())
        
        pressure_exponent = B.query_node_attribute('pressure_exponent')
        assert((pressure_exponent[wn.junction_name_list] == 0.45).all())
        
    def test_gis_roundtrip(self):
        for inp_file in self.inp_files:
            wn = self.wntr.network.WaterNetworkModel(inp_file)
            A = wn.to_gis()
            B = self.wntr.network.from_gis(A)
            assert(wn._compare(B, level=0))
    
    def test_gis_point_roundtrip(self):
        for inp_file in self.inp_files:
            wn = self.wntr.network.WaterNetworkModel(inp_file)
            A = wn.to_gis(pumps_as_points=True, valves_as_points=True)
            B = self.wntr.network.from_gis(A)
            assert(wn._compare(B, level=0))
    
    def test_gis_append_roundtrip(self):
        for inp_file in self.inp_files:
            wn = self.wntr.network.WaterNetworkModel(inp_file)
            A = wn.to_gis(pumps_as_points=True, valves_as_points=True)
            B = self.wntr.network.from_gis({'junctions': A.junctions})
            B = self.wntr.network.from_gis({'tanks': A.tanks,
                                            'reservoirs': A.reservoirs}, B)
            B = self.wntr.network.from_gis({'pipes': A.pipes}, B)
            B = self.wntr.network.from_gis({'pumps': A.pumps,
                                            'valves': A.valves}, B)
            assert(wn._compare(B, level=0))
    
    def test_geojson_roundtrip(self):
        for inp_file in self.inp_files:
            wn = self.wntr.network.WaterNetworkModel(inp_file)
            wntr.network.write_geojson(wn, 'temp', pumps_as_points=True, valves_as_points=False)
            files = {}
            if wn.num_junctions > 0:
                files['junctions'] = 'temp_junctions.geojson'
            if wn.num_tanks > 0:
                files['tanks'] = 'temp_tanks.geojson'
            if wn.num_reservoirs > 0:
                files['reservoirs'] = 'temp_reservoirs.geojson'
            if wn.num_pipes > 0:
                files['pipes'] = 'temp_pipes.geojson'
            if wn.num_pumps > 0:
                files['pumps'] = 'temp_pumps.geojson'
            if wn.num_valves > 0:
                files['valves'] = 'temp_valves.geojson'
            B = self.wntr.network.read_geojson(files)
            assert(wn._compare(B, level=0))

    def test_shapefile_roundtrip(self):
        for inp_file in self.inp_files:
            wn = self.wntr.network.WaterNetworkModel(inp_file)
            wntr.network.write_shapefile(wn, 'temp')
            files = {}
            if wn.num_junctions > 0:
                files['junctions'] = 'temp_junctions'
            if wn.num_tanks > 0:
                files['tanks'] = 'temp_tanks'
            if wn.num_reservoirs > 0:
                files['reservoirs'] = 'temp_reservoirs'
            if wn.num_tanks > 0:
                files['pipes'] = 'temp_pipes'
            if wn.num_pumps > 0:
                files['pumps'] = 'temp_pumps'
            if wn.num_valves > 0:
                files['valves'] = 'temp_valves'
            B = self.wntr.network.read_shapefile(files)
            assert(wn._compare(B, level=0))
    
    def test_valid_gis_names(self):
        
        required_names = wntr.network.io.valid_gis_names(complete_list=False, truncate_names=None)
        valid_names = wntr.network.io.valid_gis_names(complete_list=True, truncate_names=None)
        
        wn = self.wntr.network.WaterNetworkModel(join(ex_datadir, "Net6.inp"))
        gis_data = wn.to_gis()
        
        for component in required_names.keys():
            required_columns = required_names[component]
            valid_columns = valid_names[component]
            
            data = getattr(gis_data, component)
            data_columns = list(data.columns)
            data_columns.append(data.index.name)
            
            # Check that all data columns are valid
            assert len(set(data_columns)-set(valid_columns)) == 0
            # Check that all required columns are in the data
            assert len(set(required_columns)-set(data_columns)) == 0
            # Assert that node_type and link_type are not in data columns
            assert 'node_type' not in data_columns
            assert 'link_type' not in data_columns
            
if __name__ == "__main__":
    unittest.main()



# This is a test to ensure all of the examples run.
import os
import sys
import unittest
import pytest
from os import listdir
from os.path import abspath, dirname, isfile, join
from subprocess import call

testdir = dirname(abspath(str(__file__)))
examplesdir = join(testdir, "..", "..", "examples")


class TestExamples(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        import wntr

        self.wntr = wntr

    @classmethod
    def tearDownClass(self):
        pass
    
    @pytest.mark.time_consuming
    def test_that_examples_run(self):
        cwd = os.getcwd()
        os.chdir(examplesdir)
        example_files = [
            f
            for f in listdir(examplesdir)
            if isfile(join(examplesdir, f))
            and f.endswith(".py")
            and not f.startswith("test")
        ]
        flag = 0
        failed_examples = []
        for f in example_files:
            if "sensor_placement" in f:
                try:
                    import chama
                except ImportError:
                    continue
            tmp_flag = call([sys.executable, join(examplesdir, f)])
            print(f, tmp_flag)
            if tmp_flag == 1:
                failed_examples.append(f)
                flag = 1
        os.chdir(cwd)
        if len(failed_examples) > 0:
            print("failed examples: {0}".format(failed_examples))
        self.assertEqual(flag, 0)


if __name__ == "__main__":
    unittest.main()
